# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Gaussian noise agent."""

from typing import Optional, Tuple, Union

import numpy as np
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.agents.deterministic_agent import DeterministicAgent
from actorch.agents.stochastic_agent import StochasticAgent
from actorch.registry import register


__all__ = [
    "GaussianNoiseAgent",
]


@register
class GaussianNoiseAgent(StochasticAgent, DeterministicAgent):
    """Agent that adds Gaussian noise to the action."""

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        mean: "float" = 0.0,
        stddev: "float" = 0.1,
        num_random_timesteps: "int" = 0,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        mean:
            The noise Gaussian distribution mean
            (`mu` in the literature).
        stddev:
            The noise Gaussian distribution standard deviation
            (`sigma` in the literature).
        num_random_timesteps:
            The number of initial timesteps for which
            a random prediction is returned.
        device:
            The device.

        Raises
        ------
        ValueError
            If `stddev` is not in the interval (0, inf).

        """
        if stddev <= 0.0:
            raise ValueError(f"`stddev` ({stddev}) must be in the interval (0, inf)")
        self.mean = mean
        self.stddev = stddev
        StochasticAgent.__init__(
            self,
            policy,
            observation_space,
            action_space,
            is_batched,
            num_random_timesteps,
            device,
        )

    # override
    def _stochastic_predict(
        self, flat_observation: "ndarray"
    ) -> "Tuple[ndarray, ndarray]":
        flat_action, log_prob = DeterministicAgent._predict(self, flat_observation)
        flat_action += np.random.normal(
            self.mean,
            self.stddev,
            flat_action.shape,
        )
        return flat_action, log_prob

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"is_batched: {self.is_batched}, "
            f"mean: {self.mean}, "
            f"stddev: {self.stddev}, "
            f"num_random_timesteps: {self.num_random_timesteps}, "
            f"device: {self.device})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Tensor utilities."""

import math
from functools import reduce
from typing import Sequence, Union

import torch
from torch import Tensor


__all__ = [
    "base2dec",
    "broadcast_cat",
    "dec2base",
    "is_identical",
]


def broadcast_cat(
    inputs: "Sequence[Tensor]",
    dim: "int" = 0,
) -> "Tensor":
    """Concatenate a sequence of tensors along a dimension.

    Tensor shapes are expanded to the right, and broadcasting
    is applied along all dimensions except for the `dim`-th.

    Parameters
    ----------
    inputs:
        The tensors.
    dim:
        The dimension.

    Returns
    -------
        The broadcast concatenation of `inputs`.

    Raises
    ------
    IndexError
        If `dim` is out of range or not integer.

    """
    max_ndims = max([x.ndim for x in inputs])
    if dim < -max_ndims or dim >= max_ndims or not float(dim).is_integer():
        raise IndexError(
            f"`dim` ({dim}) must be in the integer interval [-{max_ndims}, {max_ndims})"
        )
    dim = (int(dim) + max_ndims) % max_ndims
    expanded_shapes = [x.shape + (1,) * (max_ndims - x.ndim) for x in inputs]
    # Broadcast along all dimensions except for the dim-th
    target_shape = [
        reduce(lambda x, y: x * y // math.gcd(x, y), sizes) if i != dim else -1
        for i, sizes in enumerate(zip(*expanded_shapes))
    ]
    return torch.cat(
        [
            x[(...,) + (None,) * (len(target_shape) - x.ndim)].repeat(
                [
                    target_size // expanded_size if target_size != -1 else 1
                    for expanded_size, target_size in zip(
                        expanded_shapes[i], target_shape
                    )
                ]
            )
            for i, x in enumerate(inputs)
        ],
        dim=dim,
    )


def is_identical(input: "Tensor", dim: "int" = 0) -> "bool":
    """Check if a tensor is identical along a dimension.

    Parameters
    ----------
    input:
        The tensor.
    dim:
        The dimension.

    Returns
    -------
        True if the tensor is identical along
        the `dim`-th dimension, False otherwise.

    Raises
    ------
    IndexError
        If `dim` is out of range or not integer.

    """
    if dim < -input.ndim or dim >= input.ndim or not float(dim).is_integer():
        raise IndexError(
            f"`dim` ({dim}) must be in the integer interval [-{input.ndim}, {input.ndim})"
        )
    dim = int(dim)
    return (input == input.select(dim, 0).unsqueeze(dim)).all(dim=dim).all()


def dec2base(
    input: "Tensor",
    base: "Union[int, Sequence[int], Tensor]" = 2,
) -> "Tensor":
    """Convert a tensor from base-10 to a given
    (possibly mixed) base.

    Parameters
    ----------
    input:
        The tensor.
    base:
        The (possibly mixed) base.

    Returns
    -------
        The tensor in the given base.

    Raises
    ------
    ValueError
        If an invalid argument value is provided.
    OverflowError
        If `input` is too large to be represented
        in the given base.

    """
    input = torch.atleast_1d(input)
    if ((input < 0) | (input != input.int())).any():
        raise ValueError(f"`input` ({input}) must be in the integer interval [0, inf)")
    base = torch.as_tensor(base)
    if ((base < 1) | (base != base.int())).any():
        raise ValueError(f"`base` ({base}) must be in the integer interval [1, inf)")
    if base.ndim > input.ndim + 1:
        raise ValueError(
            f"`base.ndim` ({base.ndim}) must be at most "
            f"equal to `input.ndim + 1` ({input.ndim + 1})"
        )
    base = torch.atleast_1d(base)
    base = base[(None,) * (input.ndim + 1 - base.ndim) + (...,)].expand(
        *input.shape[: input.ndim + 1 - base.ndim], *base.shape
    )
    n = base.prod(dim=-1)
    if (input >= n).any():
        # Overflow
        if base.shape[-1] == 1 and (base[..., 0] != 1).all():
            num_digits = (input.log() / base[..., 0].log()).floor() + 1
            base = base.repeat_interleave(int(num_digits.max()), dim=-1)
        else:
            raise OverflowError(
                f"`input` ({input}) must be in the integer "
                f"interval [{torch.zeros_like(n)}, {n})"
            )
    weights = base.flip(dims=(-1,)).roll(1, dims=-1)
    weights[..., 0] = 1
    weights = weights.cumprod(dim=-1).flip(dims=(-1,)).movedim(-1, 0)
    digits = torch.zeros((*input.shape, weights.shape[0] + 1))
    digits[..., -1] = input
    for i, weight in enumerate(weights):
        quotient = digits[..., -(i + 1)].div(weight, rounding_mode="trunc")
        remainder = digits[..., -(i + 1)] - (weight * quotient)
        digits[..., -(i + 1)], digits[..., -(i + 2)] = quotient, remainder
    return digits[..., 1:].flip(dims=(-1,)).int()


def base2dec(
    input: "Tensor",
    base: "Union[int, Sequence[int], Tensor]" = 2,
) -> "Tensor":
    """Convert a tensor from a given
    (possibly mixed) base to base-10.

    Parameters
    ----------
    input:
        The tensor.
    base:
        The (possibly mixed) base.

    Returns
    -------
        The tensor in base-10.

    Raises
    ------
    ValueError
        If an invalid argument value is provided.
    OverflowError
        If `input` is too large to be represented
        in the given base.

    """
    input = torch.atleast_1d(input)
    if ((input < 0) | (input != input.int())).any():
        raise ValueError(f"`input` ({input}) must be in the integer interval [0, inf)")
    base = torch.as_tensor(base)
    if ((base < 1) | (base != base.int())).any():
        raise ValueError(f"`base` ({base}) must be in the integer interval [1, inf)")
    try:
        base = base.expand_as(input)
    except Exception:
        raise ValueError(
            f"`base` ({base}) must be broadcastable "
            f"to the shape of `input` ({input.shape})"
        )
    if (input >= base).any():
        raise OverflowError(
            f"`input` ({input}) must be in the integer "
            f"interval [{torch.zeros_like(base)}, {base})"
        )
    weights = base.flip(dims=(-1,)).roll(1, dims=-1)
    weights[..., 0] = 1
    weights = weights.cumprod(dim=-1).flip(dims=(-1,))
    return (weights * input).sum(dim=-1)



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Experience replay buffer."""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Union

import numpy as np
from numpy import ndarray

from actorch.schedules import Schedule


__all__ = [
    "Buffer",
]


class Buffer(ABC):
    """Replay buffer that stores and samples batched experience trajectories."""

    def __init__(
        self,
        capacity: "Union[int, float]",
        spec: "Dict[str, Dict[str, Any]]",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        capacity:
            The maximum number of experiences to store. When the
            buffer overflows, old experiences are removed (FIFO).
            Set to ``float("inf")`` for unlimited capacity.
        spec:
            The specification, i.e. a dict that maps names of the values
            to store to dicts with the following key-value pairs:
            - "shape": the shape of the value to store. Default to ``()``;
            - "dtype": the dtype of the value to store. Default to ``np.float32``.

        Raises
        ------
        ValueError
            If `capacity` is not in the integer interval [1, inf].

        """
        if capacity != float("inf"):
            if capacity < 1 or not float(capacity).is_integer():
                raise ValueError(
                    f"`capacity` ({capacity}) must be in the integer interval [1, inf]"
                )
            capacity = int(capacity)
        self.capacity = capacity
        self.spec = {
            k: {
                "shape": v.get("shape", ()),
                "dtype": v.get("dtype", np.float32),
            }
            for k, v in spec.items()
        }
        self.reset()

    @property
    def schedules(self) -> "List[Schedule]":
        """Return the buffer public schedules.

        Returns
        -------
            The buffer public schedules.

        """
        return []

    def state_dict(self) -> "Dict[str, Any]":
        """Return the buffer state dict.

        Returns
        -------
            The buffer state dict.

        """
        return {
            k: v.state_dict() if hasattr(v, "state_dict") else v
            for k, v in self.__dict__.items()
        }

    def load_state_dict(self, state_dict: "Dict[str, Any]") -> "None":
        """Load a state dict into the buffer.

        Parameters
        ----------
        state_dict:
            The state dict.

        Raises
        ------
        KeyError
            If `state_dict` contains unknown keys.

        """
        for k, v in state_dict.items():
            if k not in self.__dict__:
                raise KeyError(f"{k}")
            if hasattr(self.__dict__[k], "load_state_dict"):
                self.__dict__[k].load_state_dict(v)
            else:
                self.__dict__[k] = v

    @property
    @abstractmethod
    def num_experiences(self) -> "int":
        """Return the number of stored experiences.

        Returns
        -------
            The number of stored experiences.

        """
        raise NotImplementedError

    @property
    @abstractmethod
    def num_full_trajectories(self) -> "int":
        """Return the number of stored full trajectories.

        Returns
        -------
            The number of stored full trajectories.

        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> "None":
        """Reset the buffer state."""
        raise NotImplementedError

    @abstractmethod
    def add(
        self,
        experience: "Dict[str, ndarray]",
        terminal: "ndarray",
    ) -> "None":
        """Add a batched experience to the buffer.

        In the following, let `B` denote the batch size.

        Parameters
        ----------
        experience:
            The batched experience, i.e. a dict whose key-value pairs
            are consistent with the given `spec` initialization argument,
            shape of ``experience[name]``: ``[B, *spec[name]["shape"]]``.
        terminal:
            The boolean array indicating which trajectories terminate
            at the current timestep (True), and which do not (False),
            shape: ``[B]``.

        """
        raise NotImplementedError

    @abstractmethod
    def sample(
        self,
        batch_size: "int",
        trajectory_length: "Union[int, float]" = 1,
    ) -> "Tuple[Dict[str, ndarray], ndarray, ndarray, ndarray]":
        """Sample batched experience trajectories from the buffer.

        Parameters
        ----------
        batch_size:
            The batch size.
        trajectory_length:
            The number of experiences to sample from each trajectory.
            Set to ``float("inf")`` to sample full trajectories.

        Returns
        -------
            - The sampled batched experiences, i.e. a dict whose key-value pairs
              are consistent with the given `spec` initialization argument, shape of
              ``experiences[name]``: ``[batch_size, trajectory_length, *spec[name]["shape"]]``;
            - the sampled batched indices, shape ``[batch_size, trajectory_length]``;
            - the batched importance sampling weight, shape ``[batch_size]``;
            - the mask, i.e. the boolean array indicating which trajectory
              elements are valid (True) and which are not (False),
              shape ``[batch_size, trajectory_length]``.

        """
        raise NotImplementedError

    def update_priority(self, idx: "ndarray", priority: "ndarray") -> "None":
        """Update the `idx`-th batched priority.

        In the following, let `B` denote the batch size.

        Parameters
        ----------
        idx:
            The batched index, shape: ``[B]``.
        priority:
            The new batched priority, shape: ``[B]``.

        """
        pass

    def __len__(self) -> "int":
        return self.num_experiences

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(capacity: {self.capacity}, "
            f"num_experiences: {self.num_experiences}, "
            f"num_full_trajectories: {self.num_full_trajectories})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Schedule."""

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Sequence, Union

import numpy as np
from numpy import ndarray


__all__ = [
    "Schedule",
]


class Schedule(ABC):
    """Time-dependent scheduling schema.

    Useful, for example, to anneal `epsilon` and `beta` parameters in
    epsilon-greedy exploration and prioritized experience replay, respectively.

    """

    def __init__(
        self,
        batch_size: "Optional[int]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        batch_size:
            The batch size.

        Raises
        ------
        ValueError
            If `batch_size` is not in the integer interval [1, inf).

        """
        if batch_size:
            if batch_size < 1 or not float(batch_size).is_integer():
                raise ValueError(
                    f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
                )
            batch_size = int(batch_size)
        self.batch_size = batch_size
        self.reset()

    def reset(
        self,
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
    ) -> "None":
        """Reset the (possibly batched) schedule state.

        In the following, let `B` denote the batch size.

        Parameters
        ----------
        mask:
            The boolean array indicating which batch elements are
            to reset (True) and which are not (False), shape: ``[B]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(B, dtype=bool)``.

        Raises
        ------
        ValueError
            If length of `mask` is not equal to the batch size.

        """
        batch_size = self.batch_size or 1
        if mask is None:
            mask = np.ones(batch_size, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (batch_size,))
            if len(mask) != batch_size:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be "
                    f"equal to the batch size ({batch_size})"
                )
        return self._reset(mask)

    def step(
        self,
        mask: "Optional[Union[bool, Sequence[bool], ndarray]]" = None,
    ) -> "None":
        """Step the (possibly batched) schedule state.

        In the following, let `B` denote the batch size.

        Parameters
        ----------
        mask:
            The boolean array indicating which batch elements are
            to step (True) and which are not (False), shape: ``[B]``.
            If a scalar or a singleton, it is broadcast accordingly.
            Default to ``np.ones(B, dtype=bool)``.

        Raises
        ------
        ValueError
            If length of `mask` is not equal to the batch size.

        """
        batch_size = self.batch_size or 1
        if mask is None:
            mask = np.ones(batch_size, dtype=bool)
        else:
            mask = np.array(mask, copy=False, ndmin=1)
            if len(mask) == 1:
                mask = np.broadcast_to(mask, (batch_size,))
            if len(mask) != batch_size:
                raise ValueError(
                    f"Length of `mask` ({len(mask)}) must be "
                    f"equal to the batch size ({batch_size})"
                )
        return self._step(mask)

    def state_dict(self) -> "Dict[str, Any]":
        """Return the schedule state dict.

        Returns
        -------
            The schedule state dict.

        """
        return {
            k: v.state_dict() if hasattr(v, "state_dict") else v
            for k, v in self.__dict__.items()
        }

    def load_state_dict(self, state_dict: "Dict[str, Any]") -> "None":
        """Load a state dict into the schedule.

        Parameters
        ----------
        state_dict:
            The state dict.

        Raises
        ------
        KeyError
            If `state_dict` contains unknown keys.

        """
        for k, v in state_dict.items():
            if k not in self.__dict__:
                raise KeyError(f"{k}")
            if hasattr(self.__dict__[k], "load_state_dict"):
                self.__dict__[k].load_state_dict(v)
            else:
                self.__dict__[k] = v

    @abstractmethod
    def __call__(self) -> "Union[int, float, ndarray]":
        """Return the (possibly batched) current value depending
        on the (possibly batched) schedule state.

        Returns
        -------
            The (possibly batched) current value.

        """
        raise NotImplementedError

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__}(batch_size: {self.batch_size})"

    def _reset(self, mask: "ndarray") -> "None":
        """See documentation of method `reset`."""
        pass

    def _step(self, mask: "ndarray") -> "None":
        """See documentation of method `step`."""
        pass



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Parameter noise agent."""

from copy import deepcopy
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.agents.deterministic_agent import DeterministicAgent
from actorch.agents.stochastic_agent import StochasticAgent


__all__ = [
    "ParameterNoiseAgent",
    "policy_distance_ddpg",
    "policy_distance_dqn",
]


def policy_distance_ddpg(
    policy: "Policy",
    noisy_policy: "Policy",
) -> "float":
    """Compute the DDPG-style distance between a
    policy and a noisy policy.

    Parameters
    ----------
    policy:
        The policy.
    noisy_policy:
        The noisy policy.

    Returns
    -------
        The DDPG-style distance.

    References
    ----------
    .. [1] M. Plappert, R. Houthooft, P. Dhariwal, S. Sidor, R. Y. Chen, X. Chen, T. Asfour,
           P. Abbeel, and M. Andrychowicz. "Parameter Space Noise for Exploration".
           In: ICLR. 2018.
           URL: https://arxiv.org/abs/1706.01905

    """
    delta = policy.distribution.sample() - noisy_policy.distribution.sample()
    return (delta ** 2).mean().sqrt().item()


def policy_distance_dqn(
    policy: "Policy",
    noisy_policy: "Policy",
) -> "float":
    """Compute the DQN-style distance between a
    policy and a noisy policy.

    Parameters
    ----------
    policy:
        The policy.
    noisy_policy:
        The noisy policy.

    Returns
    -------
        The DQN-style distance.

    References
    ----------
    .. [1] M. Plappert, R. Houthooft, P. Dhariwal, S. Sidor, R. Y. Chen, X. Chen, T. Asfour,
           P. Abbeel, and M. Andrychowicz. "Parameter Space Noise for Exploration".
           In: ICLR. 2018.
           URL: https://arxiv.org/abs/1706.01905

    """
    softmax = policy.distribution.sample().softmax(dim=-1)
    noisy_softmax = noisy_policy.distribution.sample().softmax(dim=-1)
    return F.kl_div(softmax, noisy_softmax, reduction="none").sum(dim=-1).mean().item()


class ParameterNoiseAgent(StochasticAgent, DeterministicAgent):
    """Agent that adds adaptive Gaussian noise to the policy weights.

    References
    ----------
    .. [1] M. Plappert, R. Houthooft, P. Dhariwal, S. Sidor, R. Y. Chen, X. Chen, T. Asfour,
           P. Abbeel, and M. Andrychowicz. "Parameter Space Noise for Exploration".
           In: ICLR. 2018.
           URL: https://arxiv.org/abs/1706.01905

    """

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        policy_distance_fn: "Callable[[Policy, Policy], float]",
        is_batched: "bool" = False,
        initial_stddev: "float" = 0.1,
        target_stddev: "float" = 0.2,
        adaption_coeff: "float" = 1.01,
        num_random_timesteps: "int" = 0,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        policy_distance_fn:
            The function that computes the distance between a
            policy and a noisy policy. This function receives
            as arguments the policy and the noisy policy, and
            returns the corresponding distance.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        initial_stddev:
            The noise Gaussian distribution initial standard deviation
            (`sigma_0` in the literature).
        target_stddev:
            The noise Gaussian distribution target standard deviation
            (`delta` in the literature).
        adaption_coeff:
            The noise Gaussian distribution adaption coefficient
            (`alpha` in the literature).
        num_random_timesteps:
            The number of initial timesteps for which
            a random prediction is returned.
        device:
            The device.

        Raises
        ------
        ValueError
            If `initial_stddev` or `target_stddev` are not in the interval (0, inf) or
            `adaption_coeff` is not in the interval (1, inf).

        """
        if initial_stddev <= 0.0:
            raise ValueError(
                f"`initial_stddev` ({initial_stddev}) must be in the interval (0, inf)"
            )
        if target_stddev <= 0.0:
            raise ValueError(
                f"`target_stddev` ({target_stddev}) must be in the interval (0, inf)"
            )
        if adaption_coeff <= 1.0:
            raise ValueError(
                f"`adaption_coeff` ({adaption_coeff}) must be in the interval (1, inf)"
            )
        self.policy_distance_fn = policy_distance_fn
        self.initial_stddev = initial_stddev
        self.target_stddev = target_stddev
        self.adaption_coeff = adaption_coeff
        self._stddev = self.initial_stddev
        self._noises = None
        StochasticAgent.__init__(
            self,
            policy,
            observation_space,
            action_space,
            is_batched,
            num_random_timesteps,
            device,
        )

    # override
    def _reset(self, mask: "ndarray") -> "None":
        super()._reset(mask)
        if self._noises:
            self._step()
        # Sample parameter noise
        with torch.no_grad():
            self._noises = [
                torch.normal(0.0, self._stddev, param.shape, device=param.device)
                for param in self.policy.parameters()
            ]

    # override
    def _stochastic_predict(
        self, flat_observation: "ndarray"
    ) -> "Tuple[ndarray, ndarray]":
        with torch.no_grad():
            # Add parameter noise
            for param, noise in zip(self.policy.parameters(), self._noises):
                param += noise
        flat_action, log_prob = DeterministicAgent._predict(self, flat_observation)
        with torch.no_grad():
            # Remove parameter noise
            for param, noise in zip(self.policy.parameters(), self._noises):
                param -= noise
        return flat_action, log_prob

    def _step(self) -> "None":
        # For simplicity, do not sample from experience
        # replay buffer but from observation space
        flat_observation = self._flat_observation_space.sample()
        policy_state_backup = self._policy_state.clone()
        # Noisy policy
        self._policy_state[...] = 0.0
        self._stochastic_predict(flat_observation)
        noisy_policy = deepcopy(self.policy)
        # Policy
        self._noises.clear()
        self._policy_state[...] = 0.0
        self._stochastic_predict(flat_observation)
        policy = self.policy
        # Restore state
        self._policy_state = policy_state_backup
        # Compute distance
        distance = self.policy_distance_fn(policy, noisy_policy)
        if distance < self.target_stddev:
            self._stddev *= self.adaption_coeff
        else:
            self._stddev /= self.adaption_coeff

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"policy_distance_fn: {self.policy_distance_fn}, "
            f"is_batched: {self.is_batched}, "
            f"initial_stddev: {self.initial_stddev}, "
            f"target_stddev: {self.target_stddev}, "
            f"adaption_coeff: {self.adaption_coeff}, "
            f"num_random_timesteps: {self.num_random_timesteps}, "
            f"device: {self.device})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Common utilities."""

import importlib
import os
import random

import numpy as np
import torch

import actorch.resources as actorch_resources


__all__ = [
    "get_resource",
    "normalize_byte_size",
    "set_seed",
]


def get_resource(path: "str") -> "str":
    """Return the absolute path to a resource.

    Parameters
    ----------
    path:
        The path to the resource, relative
        to directory `resources`.

    Returns
    -------
        The absolute path to the resource.

    """
    subpackage_name, resource_name = f"{os.sep}{os.path.normpath(path)}".rsplit(
        os.sep, 1
    )
    package_name = actorch_resources.__name__ + subpackage_name.replace(os.sep, ".")
    if package_name.endswith("."):
        package_name = package_name[:-1]
    with importlib.resources.path(package_name, resource_name) as realpath:
        return str(realpath)


def normalize_byte_size(byte_size: "int") -> "str":
    """Convert a size from B (bytes) to X, where X is the least multiple
    of B such that the normalized size is less than 1024 X.

    Parameters
    ----------
    byte_size:
        The size in bytes.

    Returns
    -------
        The normalized size.

    Raises
    ------
    ValueError
        If `byte_size` is not in the integer interval [0, inf).

    """
    if byte_size < 0 or not float(byte_size).is_integer():
        raise ValueError(
            f"`byte_size` ({byte_size}) must be in the integer interval [0, inf)"
        )
    byte_size = int(byte_size)
    factor = 1024
    for prefix in ["", "K", "M", "G", "T", "P", "E", "Z", "Y"]:
        if byte_size < factor:
            break
        byte_size /= factor
    return f"{byte_size:.2f} {prefix}B"


def set_seed(seed: "int") -> "None":
    """Set the seed for generating random numbers.

    Parameters
    ----------
    seed:
        The seed.

    Raises
    ------
    ValueError
        If `seed` is not in the integer interval [0, 2^32).

    """
    try:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
    except (TypeError, ValueError):
        raise ValueError(f"`seed` ({seed}) must be in the integer interval [0, 2^32)")



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Convolutional model."""

from typing import Any, Callable, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from actorch.models.fc import FC
from actorch.models.hydra import Hydra
from actorch.registry import register


__all__ = [
    "Conv",
]


@register
class Conv(FC):
    """Convolutional model."""

    def __init__(
        self,
        in_shapes: "Dict[str, Tuple[int, ...]]",
        out_shapes: "Dict[str, Tuple[int, ...]]",
        torso_conv_configs: "Optional[List[Dict[str, Any]]]" = None,
        torso_activation_builder: "Callable[..., nn.Module]" = nn.ReLU,
        head_fc_bias: "bool" = True,
        head_activation_builder: "Callable[..., nn.Module]" = nn.Identity,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shapes:
            The input event shapes, i.e. a dict that maps names
            of the inputs to their corresponding event shapes.
        out_shapes:
            The output event shapes, i.e. a dict that maps names
            of the outputs to their corresponding event shapes.
        torso_conv_configs:
            The configurations of the torso convolutional layers.
            Default to ``[
                {"out_channels": 8, "kernel_size": 4, "bias": True},
                {"out_channels": 4, "kernel_size": 2, "bias": True},
            ]``.
        torso_activation_builder:
            The torso activation builder
            (the same for all torso convolutional layers).
        head_fc_bias:
            True to learn an additive bias in the head
            fully connected layer, False otherwise
            (the same for all heads).
        head_activation_builder:
            The head activation builder
            (the same for all heads).

        """
        self.torso_conv_configs = torso_conv_configs or [
            {"out_channels": 8, "kernel_size": 4, "bias": True},
            {"out_channels": 4, "kernel_size": 2, "bias": True},
        ]
        self.torso_activation_builder = torso_activation_builder
        self.head_fc_bias = head_fc_bias
        self.head_activation_builder = head_activation_builder
        Hydra.__init__(self, in_shapes, out_shapes)

    # override
    def _setup_torso(self, in_shape: "Tuple[int, ...]") -> "None":
        n = len(in_shape)
        if n == 2:
            conv_builder = nn.Conv1d
        elif n == 3:
            conv_builder = nn.Conv2d
        elif n == 4:
            conv_builder = nn.Conv3d
        else:
            raise NotImplementedError(
                f"Convolution operator is not implemented for {n - 1}D input"
            )
        in_channels = [in_shape[0]] + [
            config["out_channels"] for config in self.torso_conv_configs
        ]
        torso = []
        for i in range(len(self.torso_conv_configs)):
            torso += [
                conv_builder(in_channels[i], **self.torso_conv_configs[i]),
                self.torso_activation_builder(),
            ]
        self.torso = nn.Sequential(*torso)

    # override
    def _forward_torso(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Tensor, Dict[str, Tensor]]":
        input = input[mask]
        masked_output = self.torso(input)
        out_shape = masked_output.shape[1:]
        output = torch.zeros(mask.shape + out_shape, device=masked_output.device)
        output[mask] = masked_output
        return output, states



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Normalizing flow."""

from abc import ABC, abstractmethod
from typing import Tuple

from torch import Tensor
from torch.distributions import Transform
from torch.nn import Module


__all__ = [
    "NormalizingFlow",
]


class NormalizingFlow(ABC, Transform, Module):
    """Transform with learnable parameters.

    Useful to learn arbitrarily complex distributions while retaining the ease of reparameterization of simple ones.
    Derived classes must implement `_call` to allow for sampling from the corresponding transformed distribution.
    If the transform is invertible, `_inverse`, and `log_abs_det_jacobian` should be implemented to allow for
    computing the log probability of samples drawn from the corresponding transformed distribution.
    Note that, even when `_inverse` cannot be implemented (e.g. no closed-form solution exists), the log probability
    of samples drawn from the corresponding transformed distribution can still be computed by enabling caching.

    See Also
    --------
    torch.distributions.transforms.Transform

    References
    ----------
    .. [1] I. Kobyzev, S. Prince, and M. Brubaker. "Normalizing Flows: An Introduction
           and Review of Current Methods". In: IEEE Transactions on Pattern Analysis
           and Machine Intelligence. 2020, pp. 1-1.
           URL: https://arxiv.org/abs/1908.09257
    .. [2] L. Dinh, J. Sohl-Dickstein, and S. Bengio. "Density estimation using Real NVP".
           In: ICLR. 2017.
           URL: https://arxiv.org/abs/1605.08803
    .. [3] B. Mazoure, T. Doan, A. Durand, J. Pineau, and R. D. Hjelm. "Leveraging
           exploration in off-policy algorithms via normalizing flows". In: Conference on
           Robot Learning. 2020, pp. 430-444.
           URL: https://arxiv.org/abs/1905.06893
    .. [4] P. Ward, A. Smofsky, and A. Bose. "Improving Exploration in Soft-Actor-Critic
           with Normalizing Flows Policies". In: ICML. 2019.
           URL: https://arxiv.org/abs/1906.02771

    """

    is_constant_jacobian = False
    """Whether the Jacobian matrix is constant (i.e. the transform is affine)."""

    def __init__(self, in_shape: "Tuple[int, ...]", cache_size: "int" = 0) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shape:
            The input event shape.
        cache_size:
            The cache size. If 0, no caching is done. If 1, the latest
            single value is cached. Only 0 and 1 are supported.

        """
        self.in_shape = in_shape
        super().__init__(cache_size)

    def __hash__(self) -> "int":
        return Module.__hash__(self)

    def __repr__(self) -> "str":
        return f"{self.__class__.__name__}(in_shape: {self.in_shape})"

    @abstractmethod
    def _call(self, x: "Tensor") -> "Tensor":
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Deterministic distribution."""

from typing import Optional, Union

import torch
from torch import Size, Tensor
from torch.distributions import constraints

from actorch.distributions.finite import Finite
from actorch.registry import register


__all__ = [
    "Deterministic",
]


@register
class Deterministic(Finite):
    """Distribution that returns a single deterministic value with probability equal to 1."""

    has_rsample = True
    arg_constraints = {
        "value": constraints.real,
    }

    def __init__(
        self,
        value: "Union[int, float, Tensor]",
        validate_args: "Optional[bool]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        value:
            The deterministic value to return.
        validate_args:
            True to validate the arguments, False otherwise.
            Default to ``__debug__``.

        """
        self.value = torch.as_tensor(value)
        atoms = self.value[..., None]
        probs = torch.ones_like(atoms)
        super().__init__(probs, atoms=atoms, validate_args=validate_args)

    # override
    def expand(
        self,
        batch_shape: "Size" = torch.Size(),
        _instance: "Optional[Deterministic]" = None,
    ) -> "Deterministic":
        new = self._get_checked_instance(Deterministic, _instance)
        new.value = self.value.expand(batch_shape)
        atoms = new.value[..., None]
        probs = torch.ones_like(atoms)
        super(Deterministic, new).__init__(probs, atoms=atoms, validate_args=False)
        new._validate_args = self._validate_args
        return new

    # override
    @property
    def mean(self) -> "Tensor":
        return self.value

    # override
    @property
    def variance(self) -> "Tensor":
        return torch.zeros_like(self.value)

    # override
    def sample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        with torch.no_grad():
            return self.rsample(sample_shape)

    # override
    def rsample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        shape = self._extended_shape(sample_shape)
        return self.value.expand(shape)

    # override
    def log_prob(self, value: "Tensor") -> "Tensor":
        if self._validate_args:
            self._validate_sample(value)
        expanded_value = self.value.expand_as(value)
        return (expanded_value == value).type(value.type()).log()

    # override
    def cdf(self, value: "Tensor") -> "Tensor":
        if self._validate_args:
            self._validate_sample(value)
        expanded_value = self.value.expand_as(value)
        return (expanded_value <= value).type(value.type())

    # override
    def enumerate_support(self, expand: "bool" = True) -> "Tensor":
        try:
            return super().enumerate_support(expand)
        except NotImplementedError:
            raise NotImplementedError(
                "`enumerate_support` does not support inhomogeneous values"
            )

    # override
    def entropy(self) -> "Tensor":
        return torch.zeros_like(self.value)

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(value: {self.value if self.value.numel() == 1 else self.value.shape})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Deep reinforcement learning framework for fast prototyping based on PyTorch."""

from actorch import (
    agents,
    algorithms,
    buffers,
    distributions,
    envs,
    models,
    optimizers,
    registry,
    runner,
    samplers,
    schedules,
    utils,
    visualizer,
)
from actorch.registry import register
from actorch.utils import set_seed
from actorch.version import VERSION as __version__



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Generalized estimator."""

import warnings
from contextlib import contextmanager
from typing import Iterator, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.distributions import AffineTransform, Distribution

from actorch.distributions import CatDistribution, SumTransform, TransformedDistribution


__all__ = [
    "generalized_estimator",
]


def generalized_estimator(
    state_values: "Union[Tensor, Distribution]",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    trace_weights: "Tensor",
    delta_weights: "Tensor",
    advantage_weights: "Tensor",
    action_values: "Optional[Union[Tensor, Distribution]]" = None,
    next_state_values: "Optional[Union[Tensor, Distribution]]" = None,
    discount: "float" = 0.99,
    num_return_steps: "int" = 1,
    trace_decay: "float" = 1.0,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tuple[Union[Tensor, Distribution], Tensor]":
    """Compute the (distributional) generalized estimator targets and the corresponding
    advantages of a trajectory.

    In the following, let `B` denote the batch size and `T` the trajectory maximum length.

    Parameters
    ----------
    state_values:
        The (distributional) state values (`v_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    rewards:
        The rewards (`r_t` in the literature), shape: ``[B, T]``.
    dones:
        The end-of-episode flags, shape: ``[B, T]``.
    mask:
        The boolean tensor indicating which elements (or batch elements
        if distributional) are valid (True) and which are not (False),
        shape: ``[B, T]``.
    trace_weights:
        The trace weights (`c_t` or `traces` in the literature), shape: ``[B, T]``.
    delta_weights:
        The delta weights (`rho_t` in the literature), shape: ``[B, T]``.
    advantage_weights:
        The advantage weights, shape: ``[B, T]``.
    action_values:
        The (distributional) action values (`q_t` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
    next_state_values:
        The (distributional) next state values (`v_t+1` in the literature),
        shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``.
        Default to `state_values` from the second timestep on, with a bootstrap value set to 0.
    discount:
        The discount factor (`gamma` in the literature).
    num_return_steps:
        The number of return steps (`n` in the literature).
    trace_decay:
        The trace-decay parameter (`lambda` in the literature).
    standardize_advantage:
        True to standardize the advantages, False otherwise.
    epsilon:
        The term added to the denominators to improve numerical stability.

    Returns
    -------
        - The (distributional) generalized estimator targets,
          shape (or batch shape if distributional, assuming an empty event shape): ``[B, T]``;
        - the corresponding advantages, ``[B, T]``.

    Raises
    ------
    ValueError
        If an invalid argument value is provided.

    """
    if discount < 0.0 or discount > 1.0:
        raise ValueError(f"`discount` ({discount}) must be in the interval [0, 1]")
    if num_return_steps < 1 or not float(num_return_steps).is_integer():
        raise ValueError(
            f"`num_return_steps` ({num_return_steps}) must be in the integer interval [1, inf)"
        )
    if trace_decay < 0.0 or trace_decay > 1.0:
        raise ValueError(
            f"`trace_decay` ({trace_decay}) must be in the interval [0, 1]"
        )
    if epsilon < 0.0:
        raise ValueError(f"`epsilon` ({epsilon}) must be in the interval [0, inf)")
    is_distributional = isinstance(state_values, Distribution)
    targets = (
        _compute_distributional_targets if is_distributional else _compute_targets
    )(
        state_values,
        rewards,
        dones,
        mask,
        trace_weights,
        delta_weights,
        action_values,
        next_state_values,
        discount,
        num_return_steps,
    )
    advantages = _compute_advantages(
        targets.mean if is_distributional else targets,
        state_values.mean if is_distributional else state_values,
        rewards,
        mask,
        advantage_weights,
        action_values.mean
        if (is_distributional and action_values is not None)
        else action_values,
        next_state_values.mean
        if (is_distributional and next_state_values is not None)
        else next_state_values,
        discount,
        trace_decay,
        standardize_advantage,
        epsilon,
    )
    return targets, advantages


def _compute_targets(
    state_values: "Tensor",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    trace_weights: "Tensor",
    delta_weights: "Tensor",
    action_values: "Optional[Tensor]" = None,
    next_state_values: "Optional[Tensor]" = None,
    discount: "float" = 0.99,
    num_return_steps: "int" = 1,
) -> "Tensor":
    # x[~mask] = 0.0 is equivalent to x *= mask
    # x[~mask] = 1.0 is equivalent to x += ~mask
    state_values, rewards, dones = state_values.clone(), rewards.clone(), dones.clone()
    trace_weights, delta_weights = trace_weights.clone(), delta_weights.clone()
    dones, mask = dones.bool(), mask.bool()
    state_values *= mask
    rewards *= mask
    if action_values is not None:
        action_values = action_values.clone()
        action_values *= mask
    dones += ~mask
    trace_weights += ~mask
    delta_weights += ~mask
    B, T = state_values.shape
    num_return_steps = min(num_return_steps, T)
    length = mask.sum(dim=1)
    if next_state_values is None:
        next_state_values = F.pad(state_values[:, 1:], [0, 1])
        next_state_values[torch.arange(B), length - 1] = 0.0
    else:
        next_state_values = next_state_values.clone()
    next_state_values *= mask
    next_state_values *= ~dones
    state_or_action_values = (
        action_values if action_values is not None else state_values
    )
    deltas = delta_weights * (
        rewards + discount * next_state_values - state_or_action_values
    )
    deltas *= mask
    window = discount * trace_weights
    window += ~mask
    # If the window is constant, use convolution, which is more memory and time-efficient
    is_constant_window = (window == window[:, :1])[mask].all()
    deltas = F.pad(deltas, [0, num_return_steps - 1])
    window = F.pad(window, [0, num_return_steps - 1], value=1.0)
    if not is_constant_window:
        deltas = F.unfold(deltas[:, None, :, None], (T, 1))
        window = F.unfold(window[:, None, :, None], (T, 1))
    window = window.roll(1, dims=-1)
    window[..., 0] = 1
    window = window.cumprod(dim=-1)
    errors = (
        F.conv1d(deltas[None], window[:, None, :num_return_steps], groups=B)[0]
        if is_constant_window
        else (deltas * window).sum(dim=-1)
    )
    targets = state_or_action_values + errors
    targets *= mask
    return targets


def _compute_distributional_targets(
    state_values: "Distribution",
    rewards: "Tensor",
    dones: "Tensor",
    mask: "Tensor",
    trace_weights: "Tensor",
    delta_weights: "Tensor",
    action_values: "Optional[Distribution]" = None,
    next_state_values: "Optional[Distribution]" = None,
    discount: "float" = 0.99,
    num_return_steps: "int" = 1,
) -> "Distribution":
    warnings.warn(
        "Distributional value estimation is experimental and may not work as expected",
    )
    # x[~mask] = 0.0 is equivalent to x *= mask
    # x[~mask] = 1.0 is equivalent to x += ~mask
    rewards, dones = rewards.clone(), dones.clone()
    trace_weights, delta_weights = trace_weights.clone(), delta_weights.clone()
    dones, mask = dones.bool(), mask.bool()
    rewards *= mask
    dones += ~mask
    trace_weights += ~mask
    delta_weights += ~mask
    B, T = state_values.batch_shape
    num_return_steps = min(num_return_steps, T)
    state_or_action_values = (
        action_values if action_values is not None else state_values
    )
    coeffs = torch.stack(
        [
            delta_weights * rewards,  # Offsets
            -delta_weights,  # `state_or_action_values` coefficients
            discount * delta_weights,  # `next_state_values` coefficients
        ]
    )
    window = discount * trace_weights
    window += ~mask
    coeffs = F.pad(coeffs, [0, num_return_steps - 1])
    window = F.pad(window, [0, num_return_steps - 1], value=1.0)
    # Pad mask using edge values
    mask = F.pad(mask, [0, num_return_steps - 1])
    mask[:, T - 1 :] = mask[:, T - 1 : T]
    coeffs = F.unfold(coeffs[..., None], (T, 1)).reshape(3, B, T, -1)
    window = F.unfold(window[:, None, :, None], (T, 1))
    mask = F.unfold(mask[:, None, :, None].float(), (T, 1)).bool()
    window = window.roll(1, dims=-1)
    window[..., 0] = 1
    window = window.cumprod(dim=-1)
    coeffs *= window
    coeffs[1, :, :, 0] += 1
    coeffs[2, ...] *= ~dones[..., None]
    if next_state_values is None:
        # To fix
        next_state_values = state_values
        coeffs[2, ...] = coeffs[2, ...].roll(1, dims=-1)
        coeffs[2, ..., 0] = 0.0
    coeffs *= mask
    with _patch_expand():
        state_or_action_values = state_or_action_values.expand((B, T, num_return_steps))
        next_state_values = next_state_values.expand((B, T, num_return_steps))
    state_or_action_values = TransformedDistribution(
        state_or_action_values,
        [
            AffineTransform(coeffs[0], coeffs[1]),
            SumTransform((num_return_steps,)),
        ],
        validate_args=False,
    ).reduced_dist
    next_state_values = TransformedDistribution(
        next_state_values,
        [
            AffineTransform(torch.zeros_like(coeffs[2]), coeffs[2]),
            SumTransform((num_return_steps,)),
        ],
        validate_args=False,
    ).reduced_dist
    targets = TransformedDistribution(
        CatDistribution(
            [state_or_action_values, next_state_values],
            dim=-1,
        ),
        SumTransform((2,)),
        validate_args=False,
    ).reduced_dist
    return targets


def _compute_advantages(
    targets: "Tensor",
    state_values: "Tensor",
    rewards: "Tensor",
    mask: "Tensor",
    advantage_weights: "Tensor",
    action_values: "Optional[Tensor]" = None,
    next_state_values: "Optional[Tensor]" = None,
    discount: "float" = 0.99,
    trace_decay: "float" = 1.0,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tensor":
    # x[~mask] = 0.0 is equivalent to x *= mask
    # x[~mask] = 1.0 is equivalent to x += ~mask
    targets, state_values, rewards = (
        targets.clone(),
        state_values.clone(),
        rewards.clone(),
    )
    advantage_weights = advantage_weights.clone()
    mask = mask.bool()
    targets *= mask
    state_values *= mask
    rewards *= mask
    advantage_weights += ~mask
    B, T = state_values.shape
    length = mask.sum(dim=1)
    if action_values is not None:
        # Replace action values with current action values estimate (Retrace-like)
        action_values = targets
    else:
        # Compute action values from current state values estimate (V-trace-like)
        next_targets = F.pad(
            trace_decay * targets[:, 1:] + (1 - trace_decay) * state_values[:, 1:],
            [0, 1],
        )
        if next_state_values is None:
            next_state_values = F.pad(state_values[:, 1:], [0, 1])
            next_state_values[torch.arange(B), length - 1] = 0.0
        else:
            next_state_values = next_state_values.clone()
        next_state_values *= mask
        next_targets[torch.arange(B), length - 1] = next_state_values[
            torch.arange(B), length - 1
        ]
        action_values = rewards + discount * next_targets
    advantages = advantage_weights * (action_values - state_values)
    advantages *= mask
    if standardize_advantage:
        advantages_mean = (advantages.sum(dim=1) / length)[:, None]
        advantages -= advantages_mean
        advantages *= mask
        advantages_stddev = (
            ((advantages ** 2).sum(dim=1) / length).sqrt().clamp(min=epsilon)[:, None]
        )
        advantages /= advantages_stddev
        advantages *= mask
    return advantages


def _unfoldNd(input: "Tensor", shape: "Tuple[int, ...]") -> "Tensor":
    B, T, N = shape[0:3]
    if input.ndim == 2:
        result = F.pad(input, [0, N - 1])
        return F.unfold(result[:, None, :, None], (T, 1))
    result = input.reshape(B, 1, T, 1, -1)
    result = F.pad(result, [0, 0, 0, 0, 0, N - 1]).movedim(-1, 0)
    result = torch.stack([F.unfold(x, (T, 1)) for x in result], dim=-1)
    result = result.reshape(B, T, N, *input.shape[2:])
    return result


@contextmanager
def _patch_expand() -> "Iterator[None]":
    expand_backup = torch.Tensor.expand
    try:
        torch.Tensor.expand = _unfoldNd
        yield
    finally:
        torch.Tensor.expand = expand_backup



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Serial batched environment."""

from copy import deepcopy
from typing import Any, Callable, Optional, Sequence, Tuple

import numpy as np
from gym import Env
from numpy import ndarray

from actorch.envs.batched_env import BatchedEnv
from actorch.envs.utils import Nested, batch, flatten, unflatten


__all__ = [
    "SerialBatchedEnv",
]


class SerialBatchedEnv(BatchedEnv):
    """Batched environment based on loops."""

    def __init__(
        self, env_builder: "Callable[[], Env]", num_workers: "int" = 1
    ) -> "None":
        super().__init__(env_builder, num_workers)
        dummy_flat_observation = flatten(
            self.single_observation_space,
            self.single_observation_space.sample(),
            copy=False,
        )
        self._envs = [self.env_builder() for _ in range(num_workers)]
        self._observation_buffer = [
            unflatten(
                self.single_observation_space,
                np.zeros_like(dummy_flat_observation),
                copy=False,
            )
            for _ in range(num_workers)
        ]
        self._reward = np.zeros(num_workers)
        self._done = np.zeros(num_workers, dtype=bool)
        self._info = np.array([{} for _ in range(num_workers)])

    # override
    def _reset(self, idx: "Sequence[int]") -> "Nested[ndarray]":
        for i in idx:
            self._observation_buffer[i] = self._envs[i].reset()
        observation = batch(self.observation_space, self._observation_buffer)  # Copy
        return observation

    # override
    def _step(
        self, action: "Sequence[Nested]", idx: "Sequence[int]"
    ) -> "Tuple[Nested[ndarray], ndarray, ndarray, ndarray]":
        for i in idx:
            (
                self._observation_buffer[i],
                self._reward[i],
                self._done[i],
                self._info[i],
            ) = self._envs[i].step(action[i])
        observation = batch(self.observation_space, self._observation_buffer)  # Copy
        return (
            observation,
            np.array(self._reward),  # Copy
            np.array(self._done),  # Copy
            deepcopy(self._info),
        )

    # override
    def _seed(self, seed: "Sequence[Optional[int]]") -> "None":
        for i, env in enumerate(self._envs):
            env.seed(seed[i])

    # override
    def _render(self, idx: "Sequence[int]", **kwargs: "Any") -> "None":
        for i in idx:
            self._envs[i].render(**kwargs)

    # override
    def _close(self) -> "None":
        for env in self._envs:
            env.close()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Fully connected model."""

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor, device

from actorch.models.hydra import Hydra
from actorch.registry import register


__all__ = [
    "FC",
]


@register
class FC(Hydra):
    """Fully connected model."""

    def __init__(
        self,
        in_shapes: "Dict[str, Tuple[int, ...]]",
        out_shapes: "Dict[str, Tuple[int, ...]]",
        torso_fc_configs: "Optional[List[Dict[str, Any]]]" = None,
        torso_activation_builder: "Callable[..., nn.Module]" = nn.ReLU,
        head_fc_bias: "bool" = True,
        head_activation_builder: "Callable[..., nn.Module]" = nn.Identity,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        in_shapes:
            The input event shapes, i.e. a dict that maps names
            of the inputs to their corresponding event shapes.
        out_shapes:
            The output event shapes, i.e. a dict that maps names
            of the outputs to their corresponding event shapes.
        torso_fc_configs:
            The configurations of the torso fully connected layers.
            Default to ``[
                {"out_features": 64, "bias": True},
                {"out_features": 64, "bias": True},
            ]``.
        torso_activation_builder:
            The torso activation builder
            (the same for all torso fully connected layers).
        head_fc_bias:
            True to learn an additive bias in the head
            fully connected layer, False otherwise
            (the same for all heads).
        head_activation_builder:
            The head activation builder
            (the same for all heads).

        """
        self.torso_fc_configs = torso_fc_configs or [
            {"out_features": 64, "bias": True},
            {"out_features": 64, "bias": True},
        ]
        self.torso_activation_builder = torso_activation_builder
        self.head_fc_bias = head_fc_bias
        self.head_activation_builder = head_activation_builder
        super().__init__(in_shapes, out_shapes)

    # override
    def _setup_torso(self, in_shape: "Tuple[int, ...]") -> "None":
        in_features = [torch.Size(in_shape).numel()] + [
            config["out_features"] for config in self.torso_fc_configs
        ]
        torso = []
        for i in range(len(self.torso_fc_configs)):
            torso += [
                nn.Linear(in_features[i], **self.torso_fc_configs[i]),
                self.torso_activation_builder(),
            ]
        self.torso = nn.Sequential(*torso)

    # override
    def _setup_heads(self, in_shape: "Tuple[int, ...]") -> "None":
        heads = {}
        for name, out_shape in self.out_shapes.items():
            head = [
                nn.Linear(
                    torch.Size(in_shape).numel(),
                    torch.Size(out_shape).numel(),
                    self.head_fc_bias,
                ),
                self.head_activation_builder(),
            ]
            heads[name] = nn.Sequential(*head)
        self.heads = nn.ModuleDict(heads)

    # override
    def _forward_torso(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Tensor, Dict[str, Tensor]]":
        input = input[mask].flatten(start_dim=1)
        masked_output = self.torso(input)
        out_shape = masked_output.shape[1:]
        output = torch.zeros(mask.shape + out_shape, device=masked_output.device)
        output[mask] = masked_output
        return output, states

    # override
    def _forward_heads(
        self,
        input: "Tensor",
        states: "Dict[str, Tensor]",
        mask: "Tensor",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor]]":
        outputs = {}
        for name, out_shape in self.out_shapes.items():
            masked_input = input[mask].flatten(start_dim=1)
            masked_output = self.heads[name](masked_input)
            masked_output = masked_output.reshape(-1, *out_shape)
            output = torch.zeros(mask.shape + out_shape, device=masked_output.device)
            output[mask] = masked_output
            outputs[name] = output
        return outputs, states

    # override
    def get_example_inputs(
        self,
        batch_shape: "Tuple[int, ...]" = (1,),
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "Tuple[Dict[str, Tensor], Dict[str, Tensor], Tensor]":
        inputs = {
            name: torch.rand(batch_shape + in_shape, device=device)
            for name, in_shape in self.in_shapes.items()
        }
        states = {}
        mask = torch.ones(batch_shape, dtype=torch.bool, device=device)
        return inputs, states, mask



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Masked distribution."""

from typing import Optional, Union

import torch
from torch import Size, Tensor
from torch.distributions import Distribution, constraints, utils
from torch.distributions.constraints import Constraint

from actorch.distributions.utils import is_discrete
from actorch.registry import register


__all__ = [
    "MaskedDistribution",
]


@register
class MaskedDistribution(Distribution):
    """Mask a base distribution along its batch dimensions."""

    arg_constraints = {
        "mask": constraints.boolean,
    }

    def __init__(
        self,
        base_distribution: "Distribution",
        mask: "Union[bool, Tensor]",
        validate_args: "Optional[bool]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        base_distribution:
            The base distribution.
        mask:
            The boolean scalar/tensor indicating which batch
            elements are valid (True) and which are not (False).
            Must be broadcastable to the base distribution batch shape.
        validate_args:
            True to validate the arguments, False otherwise.
            Default to ``__debug__``.

        """
        mask = torch.as_tensor(mask)
        batch_shape = torch.broadcast_shapes(base_distribution.batch_shape, mask.shape)
        if mask.shape != batch_shape:
            mask = mask.expand(batch_shape)
        if base_distribution.batch_shape != batch_shape:
            base_distribution = base_distribution.expand(batch_shape)
        self.base_dist = base_distribution
        self.mask = mask
        super().__init__(batch_shape, base_distribution.event_shape, validate_args)
        # Ensure mask validation is not bypassed
        self.mask = self.mask.bool()

    # override
    def expand(
        self,
        batch_shape: "Size" = torch.Size(),
        _instance: "Optional[MaskedDistribution]" = None,
    ) -> "MaskedDistribution":
        new = self._get_checked_instance(MaskedDistribution, _instance)
        new.base_dist = self.base_dist.expand(batch_shape)
        new.mask = self.mask.expand(batch_shape)
        super(MaskedDistribution, new).__init__(batch_shape, self.event_shape, False)
        new._validate_args = self._validate_args
        return new

    @utils.lazy_property
    def is_discrete(self) -> "bool":
        """Whether the distribution is discrete.

        Returns
        -------
            True if the distribution is discrete, False otherwise.

        """
        return is_discrete(self.base_dist)

    # override
    @property
    def support(self) -> "Constraint":
        return self.base_dist.support

    # override
    @property
    def mean(self) -> "Tensor":
        return self.base_dist.mean

    # override
    @property
    def variance(self) -> "Tensor":
        return self.base_dist.variance

    # override
    def sample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        return self.base_dist.sample(sample_shape)

    # override
    def rsample(self, sample_shape: "Size" = torch.Size()) -> "Tensor":  # noqa: B008
        return self.base_dist.rsample(sample_shape)

    # override
    def log_prob(self, value: "Tensor") -> "Tensor":
        result = self.base_dist.log_prob(value)
        return result.masked_fill(~self.mask, 0.0)

    # override
    def cdf(self, value: "Tensor") -> "Tensor":
        result = self.base_dist.cdf(value)
        return result.masked_fill(~self.mask, 0.0)

    # override
    @property
    def has_rsample(self) -> "bool":
        return self.base_dist.has_rsample

    # override
    @property
    def has_enumerate_support(self) -> "bool":
        return self.base_dist.has_enumerate_support

    # override
    def enumerate_support(self, expand: "bool" = True) -> "Tensor":
        return self.base_dist.enumerate_support(expand)

    # override
    def entropy(self) -> "Tensor":
        result = self.base_dist.entropy()
        return result.masked_fill(~self.mask, 0.0)

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"({self.base_dist}, "
            f"mask: {self.mask if self.mask.numel() == 1 else self.mask.shape})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Rank-based experience replay buffer."""

from typing import Any, Dict, Optional, Union

import numpy as np
from numpy import ndarray

from actorch.buffers.proportional_buffer import ProportionalBuffer
from actorch.registry import register
from actorch.schedules import Schedule
from actorch.utils import normalize_byte_size


__all__ = [
    "RankBasedBuffer",
]


@register
class RankBasedBuffer(ProportionalBuffer):
    """Prioritized replay buffer that samples (possibly independent)
    batched trajectories with probabilities proportional to their ranks.

    References
    ----------
    .. [1] T. Schaul, J. Quan, I. Antonoglou, and D. Silver.
           "Prioritized Experience Replay". In: ICLR. 2016.
           URL: https://arxiv.org/abs/1511.05952

    """

    def __init__(
        self,
        capacity: "Union[int, float]",
        spec: "Dict[str, Dict[str, Any]]",
        prioritization: "Optional[Schedule]" = None,
        bias_correction: "Optional[Schedule]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        capacity:
            The maximum number of experiences to store. When the
            buffer overflows, old experiences are removed (FIFO).
            Set to ``float("inf")`` for unlimited capacity.
        spec:
            The specification, i.e. a dict that maps names of the values
            to store to dicts with the following key-value pairs:
            - "shape": the shape of the value to store. Default to ``()``;
            - "dtype": the dtype of the value to store. Default to ``np.float32``.
        prioritization:
            The prioritization schedule (`alpha` in the literature;
            0: no prioritization, 1: full prioritization).
            Default to ``ConstantSchedule(1.0)``.
        bias_correction:
            The bias correction schedule (`beta` in the literature;
            0: no correction, 1: full correction).
            Default to ``ConstantSchedule(0.4)``.

        """
        super().__init__(capacity, spec, prioritization, bias_correction)
        del self.epsilon

    # override
    def add(self, experience: "Dict[str, ndarray]", terminal: "ndarray") -> "None":
        super().add(experience, terminal)
        self._priorities[self._idx] = self._max_priority

    # override
    def update_priority(self, idx: "ndarray", priority: "ndarray") -> "None":
        if ((idx >= self._num_experiences) | (idx < -self._num_experiences)).any():
            raise IndexError(
                f"`idx` ({idx}) must be in the integer interval "
                f"[-{self._num_experiences}, {self._num_experiences})"
            )
        if (priority < 0.0).any():
            raise ValueError(
                f"`priority` ({priority}) must be in the interval [0, inf)"
            )
        # Invalidate priority cache
        self._cache.pop("trajectory_priorities", None)
        self._cache.pop("full_trajectory_priority", None)
        self._priorities[idx] = priority
        self._max_priority = max(self._max_priority, priority.max())

    # override
    @property
    def _full_trajectory_priority(self) -> "ndarray":
        try:
            full_trajectory_priority = self._cache["full_trajectory_priority"]
        except KeyError:
            prioritization = self.prioritization()
            if prioritization < 0.0 or prioritization > 1.0:
                raise ValueError(
                    f"`prioritization` ({prioritization}) must be in the interval [0, 1]"
                )
            tmp = super()._full_trajectory_priority.argsort()[::-1]
            full_trajectory_rank = np.empty_like(tmp)
            full_trajectory_rank[tmp] = np.arange(self.num_full_trajectories) + 1
            full_trajectory_priority = full_trajectory_rank ** (-prioritization)
            self._cache["full_trajectory_priority"] = full_trajectory_priority
        return full_trajectory_priority

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(capacity: {self.capacity}, "
            f"prioritization: {self.prioritization}, "
            f"bias_correction: {self.bias_correction}, "
            f"num_experiences: {self.num_experiences}, "
            f"num_full_trajectories: {self.num_full_trajectories}, "
            f"size: {normalize_byte_size(self.byte_size)})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Registries."""

from actorch.distributions.registries.constraint_registry import *
from actorch.distributions.registries.kl_registry import *
from actorch.distributions.registries.reduction_registry import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""CSV loader."""

import csv
from typing import Any, Dict, Optional, Sequence, Tuple

import numpy as np
from numpy import ndarray

from actorch.visualizer.loaders.loader import Loader


__all__ = [
    "CSVLoader",
]


class CSVLoader(Loader):
    """Load progress data from CSV progress files."""

    # override
    def load(
        self,
        input_dirpath: "str",
        search_pattern: "str" = ".*",
        exclude_names: "Optional[Sequence[str]]" = None,
        **kwargs: "Any",
    ) -> "Dict[str, Dict[str, Tuple[ndarray, ndarray]]]":
        search_pattern = f"(?=.*\.csv)(?={search_pattern})"
        return super().load(input_dirpath, search_pattern, exclude_names, **kwargs)

    # override
    def _load_data(self, filepath: "str", **kwargs: "Any") -> "Dict[str, ndarray]":
        data = {}
        with open(filepath) as f:
            content = csv.reader(f)
            headers = list(next(content))
            bodies = list(zip(*[row for row in content]))
            for header, body in zip(headers, bodies):
                try:
                    # Replace empty strings with NaN
                    y_with_gaps = np.array(
                        list(map(lambda x: float(x) if not x == "" else np.nan, body))
                    )
                    if np.isfinite(y_with_gaps).any():
                        data[header] = y_with_gaps
                except ValueError:
                    pass
        return data



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Value estimation."""

from actorch.algorithms.value_estimation.generalized_estimator import *
from actorch.algorithms.value_estimation.importance_sampling import *
from actorch.algorithms.value_estimation.lambda_return import *
from actorch.algorithms.value_estimation.monte_carlo_return import *
from actorch.algorithms.value_estimation.n_step_return import *
from actorch.algorithms.value_estimation.off_policy_lambda_return import *
from actorch.algorithms.value_estimation.retrace import *
from actorch.algorithms.value_estimation.tree_backup import *
from actorch.algorithms.value_estimation.vtrace import *



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Lambda schedule."""

from typing import Callable, Optional, Union

import numpy as np
from numpy import ndarray

from actorch.schedules.schedule import Schedule


__all__ = [
    "LambdaSchedule",
]


class LambdaSchedule(Schedule):
    """Schedule defined by a given function."""

    def __init__(
        self,
        step_fn: "Callable[[Union[int, ndarray]], Union[int, float, ndarray]]",
        batch_size: "Optional[int]" = None,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        step_fn:
            The function that defines the time-dependent scheduling schema.
            This function receives as an argument the (possibly batched) number of
            elapsed timesteps, and returns the corresponding (possibly batched)
            schedule value.
        batch_size:
            The batch size.

        """
        self.step_fn = step_fn
        self._num_elapsed_timesteps = np.zeros(batch_size or 1, dtype=np.int64)
        super().__init__(batch_size)

    # override
    def _reset(self, mask: "ndarray") -> "None":
        self._num_elapsed_timesteps[mask] = -1
        self.step(mask)

    # override
    def _step(self, mask: "ndarray") -> "None":
        self._num_elapsed_timesteps[mask] += 1
        self._value = self.step_fn(self._num_elapsed_timesteps)

    def __call__(self) -> "Union[int, float, ndarray]":
        value = np.array(self._value)  # Copy
        return value if self.batch_size else value.item()

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(step_fn: {self.step_fn}, "
            f"batch_size: {self.batch_size})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Uniform experience replay buffer."""

from sys import getsizeof
from typing import Dict, Tuple, Union

import numpy as np
from numpy import ndarray

from actorch.buffers.buffer import Buffer
from actorch.buffers.utils import compute_trajectory_priorities
from actorch.registry import register
from actorch.utils import normalize_byte_size


__all__ = [
    "UniformBuffer",
]


@register
class UniformBuffer(Buffer):
    """Replay buffer that samples (possibly independent) batched experience
    trajectories uniformly at random.

    """

    _INITIAL_CAPACITY = 1024
    _CAPACITY_MULTIPLIER = 2

    # override
    @property
    def num_experiences(self) -> "int":
        return self._num_experiences

    # override
    @property
    def num_full_trajectories(self) -> "int":
        return len(self._full_trajectory_start_idx) if self._num_experiences > 0 else 0

    @property
    def byte_size(self) -> "int":
        """Return the total size in bytes of the stored experiences.

        Returns
        -------
            The total size in bytes of the stored experiences.

        """
        return self._byte_size

    # override
    def reset(self) -> "None":
        self._experiences = {
            k: np.zeros((self._INITIAL_CAPACITY, *v["shape"]), dtype=v["dtype"])
            for k, v in self.spec.items()
        }
        self._terminals = np.zeros(self._INITIAL_CAPACITY, dtype=bool)
        self._num_experiences = self._byte_size = 0
        self._idx = None
        self._overflow = False
        self._cache = {}

    # override
    def add(self, experience: "Dict[str, ndarray]", terminal: "ndarray") -> "None":
        self._cache = {}
        batch_size = len(terminal)
        if self._idx is None:
            self._idx = np.arange(-batch_size, 0, dtype=np.int64)
        if self._batch_size != batch_size:
            raise ValueError(
                f"Batch size ({batch_size}) must be constant ({self._batch_size}) "
                f"throughout calls of method `add`"
            )
        self._idx += batch_size
        if not self._overflow:
            self._num_experiences += batch_size
            self._overflow = self._num_experiences > self.capacity
        if self._overflow:
            self._num_experiences = self.capacity
            self._idx %= self.capacity
        if self._num_experiences > len(self._terminals):
            new_size = max(
                round(len(self._terminals) * self._CAPACITY_MULTIPLIER),
                self._num_experiences,
            )
            new_size = min(new_size, self.capacity)
            self._resize(new_size)
        for k, v in experience.items():
            self._experiences[k][self._idx, ...] = v
            if not self._overflow:
                self._byte_size += getsizeof(v)
        self._terminals[self._idx] = terminal

    # override
    def sample(
        self,
        batch_size: "int",
        num_timesteps: "Union[int, float]" = 1,
    ) -> "Tuple[Dict[str, ndarray], ndarray, ndarray, ndarray]":
        if batch_size < 1 or not float(batch_size).is_integer():
            raise ValueError(
                f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
            )
        if num_timesteps != float("inf"):
            if num_timesteps < 1 or not float(num_timesteps).is_integer():
                raise ValueError(
                    f"`num_timesteps` ({num_timesteps}) must be in the integer interval [1, inf]"
                )
            num_timesteps = int(num_timesteps)
        if self._num_experiences < 1:
            raise RuntimeError(
                "At least 1 experience must be added before calling method `sample`"
            )
        trajectory_start_idx, is_weight = self._compute_trajectory_start_idx(batch_size)
        idx = self._randomize_trajectory_start_idx(trajectory_start_idx, num_timesteps)
        experiences, idxes, mask = self._sample_with_idx(idx, num_timesteps)
        return experiences, idxes, is_weight, mask

    @property
    def _full_trajectory_start_idx(self) -> "ndarray":
        try:
            full_trajectory_start_idx = self._cache["full_trajectory_start_idx"]
        except KeyError:
            terminals = np.array(self._terminals)  # Copy
            terminals[self._idx] = True
            full_trajectory_start_idx = (
                np.where(terminals)[0] + self._batch_size
            ) % self._num_experiences
            self._cache["full_trajectory_start_idx"] = full_trajectory_start_idx
        return full_trajectory_start_idx

    @property
    def _trajectory_lengths(self) -> "ndarray":
        try:
            trajectory_lengths = self._cache["trajectory_lengths"]
        except KeyError:
            priorities = np.ones(self._num_experiences, dtype=np.float32)
            trajectory_lengths = compute_trajectory_priorities(
                self._idx,
                self._terminals[: self._num_experiences],
                priorities,
                self._num_experiences,
                self._batch_size,
            ).astype(np.int64)
            self._cache["trajectory_lengths"] = trajectory_lengths
        return trajectory_lengths

    @property
    def _batch_size(self) -> "int":
        return len(self._idx)

    def _resize(self, new_size: "int") -> "None":
        for k, v in self.spec.items():
            self._experiences[k].resize((new_size, *v["shape"]), refcheck=False)
        self._terminals.resize((new_size,), refcheck=False)

    def _compute_trajectory_start_idx(
        self,
        batch_size: "int",
    ) -> "Tuple[ndarray, ndarray]":
        # Select `batch_size` (possibly independent) batched
        # experience trajectories uniformly at random
        trajectory_start_idx = np.random.choice(
            self._full_trajectory_start_idx,
            batch_size,
            replace=self.num_full_trajectories < batch_size,
        )
        is_weight = np.ones(batch_size, dtype=np.float32)
        return trajectory_start_idx, is_weight

    def _randomize_trajectory_start_idx(
        self,
        trajectory_start_idx: "ndarray",
        num_timesteps: "Union[int, float]",
    ) -> "ndarray":
        # For each trajectory, select a random start index
        trajectory_length = self._trajectory_lengths[trajectory_start_idx]
        offset = self._batch_size * np.random.randint(
            0, (trajectory_length - num_timesteps).clip(1, None)
        )
        idx = trajectory_start_idx + offset
        if self._overflow:
            idx %= self.capacity
        return idx

    def _sample_with_idx(
        self,
        idx: "ndarray",
        num_timesteps: "Union[int, float]",
    ) -> "Tuple[Dict[str, ndarray], ndarray, ndarray]":
        trajectory_length = self._trajectory_lengths[idx]
        max_trajectory_length = min(num_timesteps, trajectory_length.max())
        max_trajectory_stop_idx = idx + self._batch_size * (max_trajectory_length - 1)
        idxes = np.linspace(
            idx,
            max_trajectory_stop_idx,
            max_trajectory_length,
            dtype=np.int64,
            axis=1,
        )
        mask = (
            idxes
            < (idx + self._batch_size * trajectory_length.clip(None, num_timesteps))[
                ..., None
            ]
        )
        idxes[~mask] = -1
        if self._overflow:
            idxes %= self.capacity
        experiences = {k: v[idxes] for k, v in self._experiences.items()}
        return experiences, idxes, mask

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(capacity: {self.capacity}, "
            f"num_experiences: {self.num_experiences}, "
            f"num_full_trajectories: {self.num_full_trajectories}, "
            f"size: {normalize_byte_size(self.byte_size)})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Proportional experience replay buffer."""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from actorch.buffers.uniform_buffer import UniformBuffer
from actorch.buffers.utils import compute_trajectory_priorities
from actorch.registry import register
from actorch.schedules import ConstantSchedule, Schedule
from actorch.utils import normalize_byte_size


__all__ = [
    "ProportionalBuffer",
]


@register
class ProportionalBuffer(UniformBuffer):
    """Prioritized replay buffer that samples (possibly independent) batched
    trajectories with probabilities proportional to their mean priorities.

    References
    ----------
    .. [1] T. Schaul, J. Quan, I. Antonoglou, and D. Silver.
           "Prioritized Experience Replay". In: ICLR. 2016.
           URL: https://arxiv.org/abs/1511.05952

    """

    def __init__(
        self,
        capacity: "Union[int, float]",
        spec: "Dict[str, Dict[str, Any]]",
        prioritization: "Optional[Schedule]" = None,
        bias_correction: "Optional[Schedule]" = None,
        epsilon: "float" = 1e-5,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        capacity:
            The maximum number of experiences to store. When the
            buffer overflows, old experiences are removed (FIFO).
            Set to ``float("inf")`` for unlimited capacity.
        spec:
            The specification, i.e. a dict that maps names of the values
            to store to dicts with the following key-value pairs:
            - "shape": the shape of the value to store. Default to ``()``;
            - "dtype": the dtype of the value to store. Default to ``np.float32``.
        prioritization:
            The prioritization schedule (`alpha` in the literature;
            0: no prioritization, 1: full prioritization).
            Default to ``ConstantSchedule(1.0)``.
        bias_correction:
            The bias correction schedule (`beta` in the literature;
            0: no correction, 1: full correction).
            Default to ``ConstantSchedule(0.4)``.
        epsilon:
            The term added to the priorities to ensure a
            non-zero probability of sampling each trajectory.

        Raises
        ------
        ValueError
            If an invalid argument value is provided.

        """
        if epsilon <= 0.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval (0, inf)")
        self.prioritization = prioritization or ConstantSchedule(1.0)
        self.bias_correction = bias_correction or ConstantSchedule(0.4)
        self.epsilon = epsilon
        super().__init__(capacity, spec)

    # override
    @property
    def schedules(self) -> "List[Schedule]":
        return [self.prioritization, self.bias_correction]

    # override
    def reset(self) -> "None":
        super().reset()
        self._priorities = np.zeros_like(self._terminals, dtype=np.float32)
        self._max_priority = 1.0

    # override
    def add(self, experience: "Dict[str, ndarray]", terminal: "ndarray") -> "None":
        super().add(experience, terminal)
        prioritization = self.prioritization()
        if prioritization < 0.0 or prioritization > 1.0:
            raise ValueError(
                f"`prioritization` ({prioritization}) must be in the interval [0, 1]"
            )
        self._priorities[self._idx] = self._max_priority ** prioritization

    # override
    def update_priority(self, idx: "ndarray", priority: "ndarray") -> "None":
        if ((idx >= self._num_experiences) | (idx < -self._num_experiences)).any():
            raise IndexError(
                f"`idx` ({idx}) must be in the integer interval "
                f"[-{self._num_experiences}, {self._num_experiences})"
            )
        if (priority < 0.0).any():
            raise ValueError(
                f"`priority` ({priority}) must be in the interval [0, inf)"
            )
        prioritization = self.prioritization()
        if prioritization < 0.0 or prioritization > 1.0:
            raise ValueError(
                f"`prioritization` ({prioritization}) must be in the interval [0, 1]"
            )
        # Invalidate priority cache
        self._cache.pop("trajectory_priorities", None)
        priority += self.epsilon
        self._priorities[idx] = priority ** prioritization
        self._max_priority = max(self._max_priority, priority.max())

    @property
    def _trajectory_priorities(self) -> "ndarray":
        try:
            trajectory_priorities = self._cache["trajectory_priorities"]
        except KeyError:
            trajectory_priorities = compute_trajectory_priorities(
                self._idx,
                self._terminals[: self._num_experiences],
                self._priorities[: self._num_experiences],
                self._num_experiences,
                self._batch_size,
            )
            trajectory_priorities /= self._trajectory_lengths
            self._cache["trajectory_priorities"] = trajectory_priorities
        return trajectory_priorities

    @property
    def _full_trajectory_priority(self) -> "ndarray":
        return self._trajectory_priorities[self._full_trajectory_start_idx]

    # override
    def _resize(self, new_size: "int") -> "None":
        super()._resize(new_size)
        self._priorities.resize((new_size,), refcheck=False)

    # override
    def _compute_trajectory_start_idx(
        self,
        batch_size: "int",
    ) -> "Tuple[ndarray, ndarray]":
        # Select `batch_size` (possibly independent) trajectories based on their priorities
        bias_correction = self.bias_correction()
        if bias_correction < 0.0 or bias_correction > 1.0:
            raise ValueError(
                f"`bias_correction` ({bias_correction}) must be in the interval [0, 1]"
            )
        trajectory_priority = self._full_trajectory_priority
        cdf = trajectory_priority.cumsum()
        threshold = np.random.random(batch_size) * cdf[-1]
        trajectory_idx = (np.searchsorted(cdf, threshold, side="right") - 1).clip(
            0, None
        )
        trajectory_start_idx = self._full_trajectory_start_idx[trajectory_idx]
        min_prob = cdf.min() / cdf[-1]
        prob = cdf[trajectory_idx] / cdf[-1]
        max_is_weight = (min_prob * self.num_full_trajectories) ** (-bias_correction)
        is_weight = (prob * self.num_full_trajectories) ** (
            -bias_correction
        ) / max_is_weight
        return trajectory_start_idx, is_weight

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(capacity: {self.capacity}, "
            f"prioritization: {self.prioritization}, "
            f"bias_correction: {self.bias_correction}, "
            f"epsilon: {self.epsilon}, "
            f"num_experiences: {self.num_experiences}, "
            f"num_full_trajectories: {self.num_full_trajectories}, "
            f"size: {normalize_byte_size(self.byte_size)})"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Environment utilities."""

from functools import singledispatch
from typing import Container, Dict, Hashable, List, Sequence, Tuple, TypeVar, Union

import numpy as np
from gym import spaces
from numpy import ndarray
from scipy import stats


__all__ = [
    "Nested",
    "batch",
    "batch_space",
    "flatten",
    "get_log_prob",
    "get_space_bounds",
    "nest",
    "unbatch",
    "unflatten",
    "unnest",
    "unnest_space",
]


_T = TypeVar("_T")

Nested = Union[_T, Container["Nested[_T]"]]
"""Generic nested type."""


@singledispatch
def batch_space(space: "spaces.Space", batch_size: "int" = 1) -> "spaces.Space":
    """Stack `batch_size` copies of `space`
    to form a batched space.

    Parameters
    ----------
    space:
        The space.
    batch_size:
        The batch size.

    Returns
    -------
        The batched space.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import batch_space
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @batch_space.register(Custom)
    >>> def _batch_space_custom(space, batch_size=1):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `batch_space.register`"
    )


@singledispatch
def unnest_space(
    space: "spaces.Space",
    prefix: "str" = "",
) -> "Dict[str, spaces.Space]":
    """Unnest a space, i.e. convert it to a flat
    dict that maps names of the paths to the leaf
    spaces to the leaf spaces themselves.

    Parameters
    ----------
    space:
        The space.
    prefix:
        The prefix to prepend to each
        key in the unnested space.

    Returns
    -------
        The unnested space.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import unnest_space
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @unnest_space.register(Custom)
    >>> def _unnest_space_custom(space, prefix=""):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `unnest_space.register`"
    )


@singledispatch
def get_space_bounds(space: "spaces.Space") -> "Tuple[Nested, Nested]":
    """Return the lower and upper bounds of a space.

    Parameters
    ----------
    space:
        The space.

    Returns
    -------
        - The lower bound;
        - the upper bound.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import get_space_bounds
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @get_space_bounds.register(Custom)
    >>> def _get_space_bounds_custom(space):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `get_space_bounds.register`"
    )


def flatten(
    space: "spaces.Space",
    x: "Nested",
    is_batched: "bool" = False,
    copy: "bool" = True,
    validate_args: "bool" = False,
) -> "ndarray":
    """Flatten a (possibly batched) sample
    from a (possibly batched) space.

    Parameters
    ----------
    space:
        The (possibly batched) space.
    x:
        The (possibly batched) sample.
    is_batched:
        True to preserve the batch size (if `x`
        is a batched sample from batched space
        `space`), False otherwise.
    copy:
        True to return a copy, False otherwise
        (if possible according to NumPy copying rules).
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The (possibly batched) flat sample.

    """
    unnested = unnest(space, x, validate_args)
    batch_shape = (len(unnested[0]),) if is_batched else ()
    if len(unnested) == 1:
        return np.array(unnested[0], copy=copy).reshape(*batch_shape, -1)
    arrays = []
    ndmin = 1
    for v in unnested:
        array = np.asarray(v).reshape(*batch_shape, -1)
        if array.ndim > ndmin:
            ndmin = array.ndim
        arrays.append(array)
    arrays = [np.array(v.T, copy=False, ndmin=ndmin).T for v in arrays]
    return np.concatenate(arrays, axis=-1)  # Copy


_UNFLATTEN_CACHE = {}  # Cache split_idxes to improve performance of unflatten


def unflatten(
    space: "spaces.Space",
    x: "ndarray",
    copy: "bool" = True,
    validate_args: "bool" = False,
) -> "Nested":
    """Unflatten a (possibly batched) flat sample
    from a (possibly batched) space.

    Parameters
    ----------
    space:
        The (possibly batched) space.
    x:
        The (possibly batched) flat sample.
    copy:
        True to return a copy, False otherwise
        (if possible according to NumPy copying rules).
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The (possibly batched) sample.

    """
    is_batched = x.ndim == 2
    batch_shape = (len(x),) if is_batched else ()
    try:
        split_idxes = _UNFLATTEN_CACHE[id(space), is_batched]
    except KeyError:
        unnested = unnest(space, space.sample())
        arrays = []
        ndmin = 1
        for v in unnested:
            array = np.asarray(v).reshape(*batch_shape, -1)
            if array.ndim > ndmin:
                ndmin = array.ndim
            arrays.append(array)
        lengths = np.asarray(
            [np.array(v.T, copy=False, ndmin=ndmin).T.shape[-1] for v in arrays],
            dtype=int,
        )
        split_idxes = lengths[:-1].cumsum()
        _UNFLATTEN_CACHE[id(space), is_batched] = split_idxes
    chunks = [
        np.array(v, copy=copy).reshape(*batch_shape, -1)
        for v in ([x] if len(split_idxes) == 0 else np.split(x, split_idxes, axis=-1))
    ]
    return nest(space, chunks, validate_args)


@singledispatch
def unnest(
    space: "spaces.Space",
    x: "Nested",
    validate_args: "bool" = False,
) -> "List":
    """Unnest a sample from a space, i.e. convert
    it to a flat list containing the leaf values.

    Parameters
    ----------
    space:
        The space.
    x:
        The sample.
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The unnested sample.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import unnest
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @unnest.register(Custom)
    >>> def _unnest_custom(space, x, validate_args=False):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `unnest.register`"
    )


@singledispatch
def nest(
    space: "spaces.Space",
    x: "List",
    validate_args: "bool" = False,
) -> "Nested":
    """Nest an unnested sample (i.e. a flat list
    containing the leaf values) from a space.

    Parameters
    ----------
    space:
        The space.
    x:
        The unnested sample.
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The sample.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import nest
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @nest.register(Custom)
    >>> def _nest_custom(space, x, validate_args=False):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `nest.register`"
    )


@singledispatch
def batch(
    space: "spaces.Space",
    x: "Sequence[Nested]",
    validate_args: "bool" = False,
) -> "Nested[ndarray]":
    """Batch an unbatched sample from a batched space.

    Parameters
    ----------
    space:
        The batched space.
    x:
        The unbatched sample (i.e. a sequence of samples
        from the corresponding unbatched space).
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The batched sample.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import batch
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @batch.register(Custom)
    >>> def _batch_custom(space, x, validate_args=False):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `batch.register`"
    )


@singledispatch
def unbatch(
    space: "spaces.Space",
    x: "Nested[ndarray]",
    validate_args: "bool" = False,
) -> "List[Nested]":
    """Unbatch a batched sample from a batched space.

    Parameters
    ----------
    space:
        The batched space.
    x:
        The batched sample.
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The unbatched sample (i.e. a list of samples
        from the corresponding unbatched space).

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import unbatch
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @unbatch.register(Custom)
    >>> def _unbatch_custom(space, x, validate_args=False):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `unbatch.register`"
    )


@singledispatch
def get_log_prob(
    space: "spaces.Space",
    x: "Nested",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    """Return the log probability of a (possibly batched)
    sample from a (possibly batched) space.

    Parameters
    ----------
    space:
        The (possibly batched) space.
    x:
        The (possibly batched) sample.
    is_batched:
        True to preserve the batch size (if `x`
        is a batched sample from batched space
        `space`), False otherwise.
    validate_args:
        True to validate the arguments, False otherwise.

    Returns
    -------
        The (possibly batched) log probability.

    Raises
    ------
    NotImplementedError
        If the space type is not supported.

    Notes
    -----
    Register a custom space type as follows:
    >>> from gym import spaces
    >>>
    >>> from actorch.envs.utils import get_log_prob
    >>>
    >>>
    >>> class Custom(spaces.Space):
    >>>     # Implementation
    >>>
    >>>
    >>> @unbatch.register(Custom)
    >>> def _get_log_prob_custom(space, x, is_batched=False, validate_args=False):
    >>>     # Implementation

    """
    raise NotImplementedError(
        f"Unsupported space type: {type(space)}. "
        f"Register a custom space type via "
        f"decorator `get_log_prob.register`"
    )


#####################################################################################################
# batch_space implementation
#####################################################################################################


@batch_space.register(spaces.Box)
def _batch_space_box(space: "spaces.Box", batch_size: "int" = 1) -> "spaces.Box":
    if batch_size < 1 or not float(batch_size).is_integer():
        raise ValueError(
            f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
        )
    batch_size = int(batch_size)
    reps = (batch_size,) + (1,) * len(space.shape)
    low, high = np.tile(space.low, reps), np.tile(space.high, reps)
    return spaces.Box(low, high, dtype=space.dtype)


@batch_space.register(spaces.Discrete)
def _batch_space_discrete(
    space: "spaces.Discrete", batch_size: "int" = 1
) -> "spaces.MultiDiscrete":
    if batch_size < 1 or not float(batch_size).is_integer():
        raise ValueError(
            f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
        )
    batch_size = int(batch_size)
    return spaces.MultiDiscrete(np.full((batch_size,), space.n, space.dtype))


@batch_space.register(spaces.MultiBinary)
def _batch_space_multi_binary(
    space: "spaces.MultiBinary", batch_size: "int" = 1
) -> "spaces.MultiBinary":
    if batch_size < 1 or not float(batch_size).is_integer():
        raise ValueError(
            f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
        )
    batch_size = int(batch_size)
    return spaces.MultiBinary((batch_size,) + space.shape)


@batch_space.register(spaces.MultiDiscrete)
def _batch_space_multi_discrete(
    space: "spaces.MultiDiscrete",
    batch_size: "int" = 1,
) -> "spaces.MultiDiscrete":
    if batch_size < 1 or not float(batch_size).is_integer():
        raise ValueError(
            f"`batch_size` ({batch_size}) must be in the integer interval [1, inf)"
        )
    batch_size = int(batch_size)
    reps = (batch_size,) + (1,) * len(space.shape)
    nvec = np.tile(space.nvec, reps)
    return spaces.MultiDiscrete(nvec, space.dtype)


@batch_space.register(spaces.Tuple)
def _batch_space_tuple(
    space: "spaces.Tuple",
    batch_size: "int" = 1,
) -> "spaces.Tuple":
    return spaces.Tuple([batch_space(v, batch_size) for v in space])


@batch_space.register(spaces.Dict)
def _batch_space_dict(
    space: "spaces.Dict",
    batch_size: "int" = 1,
) -> "spaces.Dict":
    return spaces.Dict({k: batch_space(v, batch_size) for k, v in space.items()})


#####################################################################################################
# unnest_space implementation
#####################################################################################################


@unnest_space.register(spaces.Box)
@unnest_space.register(spaces.Discrete)
@unnest_space.register(spaces.MultiBinary)
@unnest_space.register(spaces.MultiDiscrete)
def _unnest_space_base(
    space: "Union[spaces.Box, spaces.Discrete, spaces.MultiBinary, spaces.MultiDiscrete]",
    prefix: "str" = "",
) -> "Dict[str, Union[spaces.Box, spaces.Discrete, spaces.MultiBinary, spaces.MultiDiscrete]]":
    return {prefix: space}


@unnest_space.register(spaces.Tuple)
def _unnest_space_tuple(
    space: "spaces.Tuple",
    prefix: "str" = "",
) -> "Dict[str, spaces.Space]":
    result = {}
    for i, v in enumerate(space):
        result.update(unnest_space(v, f"{prefix}/{i}"))
    return result


@unnest_space.register(spaces.Dict)
def _unnest_space_dict(
    space: "spaces.Dict",
    prefix: "str" = "",
) -> "Dict[str, spaces.Space]":
    result = {}
    for k, v in space.items():
        result.update(unnest_space(v, f"{prefix}/{k}"))
    return result


#####################################################################################################
# get_space_bounds implementation
#####################################################################################################


@get_space_bounds.register(spaces.Box)
def _get_space_bounds_box(space: "spaces.Box") -> "Tuple[ndarray, ndarray]":
    return space.low, space.high


@get_space_bounds.register(spaces.Discrete)
def _get_space_bounds_discrete(space: "spaces.Discrete") -> "Tuple[int, int]":
    return 0, space.n - 1


@get_space_bounds.register(spaces.MultiBinary)
def _get_space_bounds_multi_binary(
    space: "spaces.MultiBinary",
) -> "Tuple[ndarray, ndarray]":
    low = np.zeros(space.shape, space.dtype)
    high = np.ones_like(low)
    return low, high


@get_space_bounds.register(spaces.MultiDiscrete)
def _get_space_bounds_multi_discrete(
    space: "spaces.MultiDiscrete",
) -> "Tuple[ndarray, ndarray]":
    high = space.nvec - 1
    low = np.zeros_like(high)
    return low, high


@get_space_bounds.register(spaces.Tuple)
def _get_space_bounds_tuple(
    space: "spaces.Tuple",
) -> "Tuple[Tuple[Nested, ...], Tuple[Nested, ...]]":
    lows, highs = [], []
    for v in space:
        low, high = get_space_bounds(v)
        lows.append(low)
        highs.append(high)
    return tuple(lows), tuple(highs)


@get_space_bounds.register(spaces.Dict)
def _get_space_bounds_dict(
    space: "spaces.Dict",
) -> "Tuple[Dict[Hashable, Nested], Dict[Hashable, Nested]]":
    lows, highs = {}, {}
    for k, v in space.items():
        lows[k], highs[k] = get_space_bounds(v)
    return lows, highs


#####################################################################################################
# unnest implementation
#####################################################################################################


@unnest.register(spaces.Box)
@unnest.register(spaces.Discrete)
@unnest.register(spaces.MultiBinary)
@unnest.register(spaces.MultiDiscrete)
def _unnest_base(
    space: "Union[spaces.Box, spaces.Discrete, spaces.MultiBinary, spaces.MultiDiscrete]",
    x: "Union[int, ndarray]",
    validate_args: "bool" = False,
) -> "List[Union[int, ndarray]]":
    if validate_args and x not in space:
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    return [x]


@unnest.register(spaces.Tuple)
def _unnest_tuple(
    space: "spaces.Tuple",
    x: "Tuple[Nested, ...]",
    validate_args: "bool" = False,
) -> "List":
    if validate_args and not (isinstance(x, tuple) and len(x) == len(space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    result = []
    for i, v in enumerate(space):
        result += unnest(v, x[i], validate_args)
    return result


@unnest.register(spaces.Dict)
def _unnest_dict(
    space: "spaces.Dict",
    x: "Dict[Hashable, Nested]",
    validate_args: "bool" = False,
) -> "List":
    if validate_args and not (isinstance(x, dict) and all(k in x for k in space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    result = []
    for k, v in space.items():
        result += unnest(v, x[k], validate_args)
    return result


#####################################################################################################
# nest implementation
#####################################################################################################


@nest.register(spaces.Box)
@nest.register(spaces.MultiBinary)
@nest.register(spaces.MultiDiscrete)
def _nest_box_multi_binary_multi_discrete(
    space: "Union[spaces.Box, spaces.MultiBinary, spaces.MultiDiscrete]",
    x: "List[ndarray]",
    validate_args: "bool" = False,
) -> "ndarray":
    try:
        result = x[0].reshape(space.shape).astype(space.dtype)
        if validate_args and result not in space:
            raise ValueError
        return result
    except Exception:
        raise ValueError(f"`x` ({x}) must be an unnested sample from `space` ({space})")


@nest.register(spaces.Discrete)
def _nest_discrete(
    space: "spaces.Discrete",
    x: "List[int]",
    validate_args: "bool" = False,
) -> "int":
    try:
        result = float(x[0])
        if not result.is_integer():
            raise ValueError
        result = int(result)
        if validate_args and result not in space:
            raise ValueError
        return result
    except Exception:
        raise ValueError(f"`x` ({x}) must be an unnested sample from `space` ({space})")


_NEST_CACHE = {}  # Cache split_idxes to improve performance of nest


@nest.register(spaces.Tuple)
def _nest_tuple(
    space: "spaces.Tuple",
    x: "List",
    validate_args: "bool" = False,
) -> "Tuple[Nested, ...]":
    try:
        split_idxes = _NEST_CACHE[id(space)]
    except KeyError:
        lengths = [len(unnest(v, v.sample())) for v in space]
        split_idxes = np.asarray([0] + lengths, dtype=int).cumsum()
        _NEST_CACHE[id(space)] = split_idxes
    try:
        result = tuple(
            nest(v, x[split_idxes[i] : split_idxes[i + 1]], validate_args)
            for i, v in enumerate(space)
        )
        return result
    except Exception:
        raise ValueError(f"`x` ({x}) must be an unnested sample from `space` ({space})")


@nest.register(spaces.Dict)
def _nest_dict(
    space: "spaces.Dict",
    x: "List",
    validate_args: "bool" = False,
) -> "Dict[Hashable, Nested]":
    try:
        split_idxes = _NEST_CACHE[id(space)]
    except KeyError:
        lengths = [len(unnest(v, v.sample())) for v in space.values()]
        split_idxes = np.asarray([0] + lengths, dtype=int).cumsum()
        _NEST_CACHE[id(space)] = split_idxes
    try:
        result = {
            k: nest(v, x[split_idxes[i] : split_idxes[i + 1]], validate_args)
            for i, (k, v) in enumerate(space.items())
        }
        return result
    except Exception:
        raise ValueError(f"`x` ({x}) must be an unnested sample from `space` ({space})")


#####################################################################################################
# batch implementation
#####################################################################################################


@batch.register(spaces.Box)
@batch.register(spaces.MultiBinary)
@batch.register(spaces.MultiDiscrete)
def _batch_base(
    space: "Union[spaces.Box, spaces.MultiBinary, spaces.MultiDiscrete]",
    x: "Sequence[ndarray]",
    validate_args: "bool" = False,
) -> "ndarray":
    try:
        result = np.stack(x)
        if validate_args and result not in space:
            raise ValueError
        return result
    except Exception:
        raise ValueError(
            f"`x` ({x}) must be an unbatched sample from `space` ({space})"
        )


@batch.register(spaces.Tuple)
def _batch_tuple(
    space: "spaces.Tuple",
    x: "Sequence[Tuple[Nested, ...]]",
    validate_args: "bool" = False,
) -> "Tuple[Nested[ndarray], ...]":
    try:
        batch_size = len(x)
        result = tuple(
            batch(v, [x[i][j] for i in range(batch_size)], validate_args)
            for j, v in enumerate(space)
        )
        if validate_args and not (
            isinstance(result, tuple) and len(result) == len(space)
        ):
            raise ValueError
        return result
    except Exception:
        raise ValueError(
            f"`x` ({x}) must be an unbatched sample from `space` ({space})"
        )


@batch.register(spaces.Dict)
def _batch_dict(
    space: "spaces.Dict",
    x: "Sequence[Dict[Hashable, Nested]]",
    validate_args: "bool" = False,
) -> "Dict[Hashable, Nested[ndarray]]":
    try:
        batch_size = len(x)
        result = {
            k: batch(v, [x[i][k] for i in range(batch_size)], validate_args)
            for k, v in space.items()
        }
        if validate_args and not (
            isinstance(result, dict) and all(k in result for k in space)
        ):
            raise ValueError
        return result
    except Exception:
        raise ValueError(
            f"`x` ({x}) must be an unbatched sample from `space` ({space})"
        )


#####################################################################################################
# unbatch implementation
#####################################################################################################


@unbatch.register(spaces.Box)
@unbatch.register(spaces.MultiBinary)
@unbatch.register(spaces.MultiDiscrete)
def _unbatch_base(
    space: "Union[spaces.Box, spaces.MultiBinary, spaces.MultiDiscrete]",
    x: "ndarray",
    validate_args: "bool" = False,
) -> "List[ndarray]":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if len(space.shape) == 0:
            raise ValueError(f"`x` ({x}) and `space` ({space}) must be batched")
    return list(x)


@unbatch.register(spaces.Tuple)
def _unbatch_tuple(
    space: "spaces.Tuple",
    x: "Tuple[Nested[ndarray], ...]",
    validate_args: "bool" = False,
) -> "List[Tuple[Nested, ...]]":
    if validate_args and not (isinstance(x, tuple) and len(x) == len(space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    result = [unbatch(v, x[i], validate_args) for i, v in enumerate(space)]
    batch_size = len(result[0])
    if validate_args and any(len(v) != batch_size for v in result[1:]):
        raise ValueError(
            f"Batch sizes ({[len(v) for v in result]}) must be "
            f"equal across samples from different subspaces"
        )
    return [tuple(result[j][i] for j in range(len(space))) for i in range(batch_size)]


@unbatch.register(spaces.Dict)
def _unbatch_dict(
    space: "spaces.Dict",
    x: "Dict[Hashable, Nested[ndarray]]",
    validate_args: "bool" = False,
) -> "List[Dict[Hashable, Nested]]":
    if validate_args and not (isinstance(x, dict) and all(k in x for k in space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    result = [unbatch(v, x[k], validate_args) for k, v in space.items()]
    batch_size = len(result[0])
    if validate_args and any(len(v) != batch_size for v in result[1:]):
        raise ValueError(
            f"Batch sizes ({[len(v) for v in result]}) must be "
            f"equal across samples from different subspaces"
        )
    return [{k: result[j][i] for j, k in enumerate(space)} for i in range(batch_size)]


#####################################################################################################
# get_log_prob implementation
#####################################################################################################


@get_log_prob.register(spaces.Box)
def _get_log_prob_box(
    space: "spaces.Box",
    x: "ndarray",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if is_batched and len(space.shape) == 0:
            raise ValueError(f"`x` ({x}) and `space` ({space}) must be batched")
    unbounded = ~space.bounded_below & ~space.bounded_above
    lower_bounded = space.bounded_below & ~space.bounded_above
    upper_bounded = ~space.bounded_below & space.bounded_above
    bounded = space.bounded_below & space.bounded_above
    result = np.empty_like(x)
    result[unbounded] = stats.norm().logpdf(x[unbounded])
    result[lower_bounded] = stats.expon().logpdf(
        x[lower_bounded] - space.low[lower_bounded]
    )
    result[upper_bounded] = stats.expon().logpdf(
        space.high[upper_bounded] - x[upper_bounded]
    )
    result[bounded] = stats.uniform(
        loc=space.low[bounded],
        scale=space.high[bounded] - space.low[bounded],
    ).logpdf(x[bounded])
    return result.sum(axis=tuple(range(is_batched, result.ndim)))


@get_log_prob.register(spaces.Discrete)
def _get_log_prob_discrete(
    space: "spaces.Discrete",
    x: "int",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if is_batched:
            raise ValueError(
                f"`is_batched` ({is_batched}) must be "
                f"False for space of type {type(space)}"
            )
    return stats.randint(0, space.n).logpmf(x)


@get_log_prob.register(spaces.MultiBinary)
def _get_log_prob_multi_binary(
    space: "spaces.MultiBinary",
    x: "ndarray",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if is_batched and len(space.shape) == 0:
            raise ValueError(f"`x` ({x}) and `space` ({space}) must be batched")
    result = np.log(np.full(x.shape, 0.5))
    return result.sum(axis=tuple(range(is_batched, result.ndim)))


@get_log_prob.register(spaces.MultiDiscrete)
def _get_log_prob_multi_discrete(
    space: "spaces.MultiDiscrete",
    x: "ndarray",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args:
        if x not in space:
            raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
        if is_batched and len(space.shape) == 0:
            raise ValueError(f"`x` ({x}) and `space` ({space}) must be batched")
    result = stats.randint(np.zeros_like(space.nvec), space.nvec).logpmf(x)
    return result.sum(axis=tuple(range(is_batched, result.ndim)))


@get_log_prob.register(spaces.Tuple)
def _get_log_prob_tuple(
    space: "spaces.Tuple",
    x: "Tuple[Nested, ...]",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args and not (isinstance(x, tuple) and len(x) == len(space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    arrays = [
        get_log_prob(v, x[i], is_batched, validate_args) for i, v in enumerate(space)
    ]
    try:
        return np.asarray(arrays).sum(axis=0)
    except Exception:
        raise ValueError(
            f"Batch sizes ({[len(v) for v in arrays]}) must be "
            f"equal across samples from different subspaces"
        )


@get_log_prob.register(spaces.Dict)
def _get_log_prob_dict(
    space: "spaces.Dict",
    x: "Dict[Hashable, Nested]",
    is_batched: "bool" = False,
    validate_args: "bool" = False,
) -> "ndarray":
    if validate_args and not (isinstance(x, dict) and all(k in x for k in space)):
        raise ValueError(f"`x` ({x}) must be a sample from `space` ({space})")
    arrays = [
        get_log_prob(v, x[k], is_batched, validate_args) for k, v in space.items()
    ]
    try:
        return np.asarray(arrays).sum(axis=0)
    except Exception:
        raise ValueError(
            f"Batch sizes ({[len(v) for v in arrays]}) must be "
            f"equal across samples from different subspaces"
        )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Conjugate gradient backtracking line search optimizer."""

# Adapted from:
# https://github.com/rlworkgroup/garage/blob/6585ac7a5f0adda75a8deda7c5fac5b92cfc0385/src/garage/torch/optimizers/conjugate_gradient_optimizer.py#L1

import logging
from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

import torch
from torch import Tensor, optim

from actorch.registry import register


__all__ = [
    "CGBLS",
]


_LOGGER = logging.getLogger(__name__)


@register
class CGBLS(optim.Optimizer):
    """Conjugate gradient backtracking line search optimizer.

    Perform constrained optimization via backtracking line search.
    The search direction is computed using the conjugate gradient method,
    which gives `x = A^(-1) g`, where `A` is a second order approximation
    of the constraint and `g` the gradient of the loss function.

    References
    ----------
    .. [1] J. Schulman, S. Levine, P. Moritz, M. Jordan, and P. Abbeel.
           "Trust Region Policy Optimization". In: ICML. 2015, pp. 1889-1897.
           URL: https://arxiv.org/abs/1502.05477

    """

    def __init__(
        self,
        params: "Union[Iterable[Tensor], Iterable[Dict[str, Any]]]",
        max_constraint: "float",
        cg_iters: "int" = 10,
        max_backtracks: "int" = 15,
        backtrack_ratio: "float" = 0.8,
        hvp_reg_coeff: "float" = 1e-5,
        accept_violation: "bool" = False,
        epsilon: "float" = 1e-8,
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        params:
            The parameters to optimize.
        max_constraint:
            The maximum constraint value.
        cg_iters:
            The number of conjugate gradient iterations for `A^(-1) g` computation.
        max_backtracks:
            The maximum number of backtracking line search iterations.
        backtrack_ratio:
            The backtrack ratio for backtracking line search.
        hvp_reg_coeff:
            The regularization coefficient for the Hessian-vector product computation.
        accept_violation:
            True to accept the descent step if it violates the line search
            condition after exhausting all backtracking budgets, False otherwise.
        epsilon:
            The term added to the denominators to improve numerical stability.

        Raises
        ------
        ValueError
            If an invalid argument value is provided.

        """
        if cg_iters < 1 or not float(cg_iters).is_integer():
            raise ValueError(
                f"`cg_iters` ({cg_iters}) must be in the integer interval [1, inf)"
            )
        if max_backtracks < 1 or not float(max_backtracks).is_integer():
            raise ValueError(
                f"`max_backtracks` ({max_backtracks}) must be in the integer interval [1, inf)"
            )
        if backtrack_ratio >= 1.0 or backtrack_ratio <= 0.0:
            raise ValueError(
                f"`backtrack_ratio` ({backtrack_ratio}) must be in the interval (0, 1)"
            )
        if hvp_reg_coeff <= 0.0:
            raise ValueError(
                f"`hvp_reg_coeff` ({hvp_reg_coeff}) must be in the interval (0, inf)"
            )
        if epsilon <= 0.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval (0, inf)")
        cg_iters = int(cg_iters)
        max_backtracks = int(max_backtracks)

        defaults = {
            "max_constraint": max_constraint,
            "cg_iters": cg_iters,
            "max_backtracks": max_backtracks,
            "backtrack_ratio": backtrack_ratio,
            "hvp_reg_coeff": hvp_reg_coeff,
            "accept_violation": accept_violation,
            "epsilon": epsilon,
        }
        super().__init__(params, defaults)

    # override
    @torch.no_grad()
    def step(
        self,
        loss_fn: "Callable[[], Tensor]",
        constraint_fn: "Callable[[], Tensor]",
    ) -> "Tuple[Tensor, Tensor]":
        """Perform a single optimization step (parameter update).

        Parameters
        ----------
        loss_fn:
            The function that computes the loss, i.e. a function that receives
            no arguments and returns the loss value.
        constraint_fn:
            The function that computes the constraint, i.e. a function that receives
            no arguments and returns the constraint value.

        Returns
        -------
            The reevaluated loss value; the reevaluated constraint value.

        """
        loss, constraint = 0.0, 0.0
        for group in self.param_groups:
            # Collect trainable parameters and gradients
            params = []
            grads = []
            for param in group["params"]:
                if param.grad is not None:
                    params.append(param)
                    grads.append(param.grad.flatten())
            flat_loss_grads = torch.cat(grads)

            # Build Hessian-vector product function
            Ax_fn = self._build_hessian_vector_product(
                constraint_fn, params, group["hvp_reg_coeff"]
            )

            # Compute step direction
            step_dir = self._conjugate_gradient(
                Ax_fn, flat_loss_grads, group["cg_iters"]
            )

            # Replace NaN with 0
            step_dir[step_dir.ne(step_dir)] = 0.0

            # Compute step size
            step_size = (
                2.0
                * group["max_constraint"]
                * (1.0 / (step_dir.dot(Ax_fn(step_dir)) + group["epsilon"]))
            ).sqrt()

            if step_size.isnan():
                step_size = 1.0

            descent_step = step_size * step_dir

            # Update parameters using backtracking line search
            loss, constraint = self._backtracking_line_search(
                params,
                descent_step,
                loss_fn,
                constraint_fn,
                group["max_constraint"],
                group["max_backtracks"],
                group["backtrack_ratio"],
                group["accept_violation"],
            )

        return loss, constraint

    def _build_hessian_vector_product(
        self,
        fn: "Callable[[], Tensor]",
        params: "List[Tensor]",
        hvp_reg_coeff: "float" = 1e-5,
    ) -> "Callable[[Tensor], Tensor]":
        """Compute the Hessian-vector product using Pearlmutter's algorithm.

        Parameters
        ----------
        fn:
            The function whose Hessian is to be computed.
        params:
            The function learnable parameters.
        hvp_reg_coeff:
            The regularization coefficient for the Hessian-vector product computation.

        Returns
        -------
            The function that computes the Hessian-vector product.
            This function receives as an argument the vector to be multiplied
            with the Hessian and returns the corresponding Hessian-vector product.

        References
        ----------
        .. [1] B. A. Pearlmutter. "Fast Exact Multiplication by the Hessian".
               In: Neural Computation. 1994, pp. 147-160.
               URL: https://doi.org/10.1162/neco.1994.6.1.147

        """
        param_shapes = [p.shape or (1,) for p in params]
        with torch.enable_grad():
            f = fn()
        f_grads = torch.autograd.grad(f, params, create_graph=True)

        def eval_fn(vector: "Tensor") -> "Tensor":
            """The evaluation function.

            Parameters
            ----------
            vector:
                The vector to be multiplied with the Hessian.

            Returns
            -------
                The product of the Hessian of `fn` and `vector`.

            """
            unflattened_vectors = [
                x.reshape(shape)
                for x, shape in zip(
                    vector.split(
                        list(map(lambda x: torch.Size(x).numel(), param_shapes))
                    ),
                    param_shapes,
                )
            ]
            assert len(f_grads) == len(unflattened_vectors)

            with torch.enable_grad():
                grad_vector_product = torch.stack(
                    [(g * x).sum() for g, x in zip(f_grads, unflattened_vectors)]
                ).sum()
            hvp = list(
                torch.autograd.grad(grad_vector_product, params, retain_graph=True)
            )
            for i, (hx, p) in enumerate(zip(hvp, params)):
                if hx is None:
                    hvp[i] = torch.zeros_like(p)

            flat_output = torch.cat([h.flatten() for h in hvp])
            return flat_output + hvp_reg_coeff * vector

        return eval_fn

    def _conjugate_gradient(
        self,
        Ax_fn: "Callable[[Tensor], Tensor]",
        b: "Tensor",
        cg_iters: "int" = 10,
        residual_tol: "float" = 1e-10,
    ) -> "Tensor":
        """Use conjugate gradient method to solve `Ax = b`.

        Parameters
        ----------
        Ax_fn:
            The function that computes the Hessian-vector product.
            This function receives as an argument the vector to be multiplied
            with the Hessian and returns the corresponding Hessian-vector product.
        b:
            Right-hand side of the equation to solve.
        cg_iters:
            The number of conjugate gradient iterations for `A^(-1) b` computation.
        residual_tol:
            The tolerance for convergence.

        Returns
        -------
            The solution `x*` for equation `Ax = b`.

        References
        ----------
        .. [1] M.R. Hestenes, and E. Stiefel. "Methods of Conjugate Gradients for
               Solving Linear Systems". In: Journal of Research of the National
               Bureau of Standards. 1952, pp. 409-435.
               URL: http://dx.doi.org/10.6028/jres.049.044

        """
        p, r = b.clone(), b.clone()
        x = torch.zeros_like(b)
        r_dot_r = r.dot(r)

        for _ in range(cg_iters):
            z = Ax_fn(p)
            v = r_dot_r / p.dot(z)
            x += v * p
            r -= v * z
            new_r_dot_r = r.dot(r)
            mu = new_r_dot_r / r_dot_r
            p = r + mu * p

            r_dot_r = new_r_dot_r
            if r_dot_r < residual_tol:
                break
        return x

    def _backtracking_line_search(
        self,
        params: "List[Tensor]",
        descent_step: "Tensor",
        loss_fn: "Callable[[], Tensor]",
        constraint_fn: "Callable[[], Tensor]",
        max_constraint: "float",
        max_backtracks: "int" = 15,
        backtrack_ratio: "float" = 0.8,
        accept_violation: "bool" = False,
    ) -> "Tuple[Tensor, Tensor]":
        """Use backtracking line search to update the parameters.

        Parameters
        ----------
        params:
            The parameters to optimize.
        descent_step:
            The descent step.
        loss_fn:
            The function that computes the loss, i.e. a function that receives
            no arguments and returns the loss value.
        constraint_fn:
            The function that computes the constraint, i.e. a function that receives
            no arguments and returns the constraint value.
        max_constraint:
            The maximum constraint value.
        max_backtracks:
            The maximum number of backtracking line search iterations.
        backtrack_ratio:
            The backtrack ratio.
        accept_violation:
            True to accept the descent step if it violates the line search
            condition after exhausting all backtracking budgets, False otherwise.

        Returns
        -------
            The reevaluated loss value; the reevaluated constraint value.

        References
        ----------
        .. [1] L. Armijo. "Minimization of functions having Lipschitz continuous
               first partial derivatives". In: Pacific J. Math. 1966.
               URL: https://msp.org/pjm/1966/16-1/p01.xhtml

        """
        prev_params = [p.clone() for p in params]
        ratios = backtrack_ratio ** torch.arange(max_backtracks)
        loss_before = loss_fn()

        param_shapes = [p.shape or (1,) for p in params]
        descent_steps = [
            x.reshape(shape)
            for x, shape in zip(
                descent_step.split(
                    list(map(lambda x: torch.Size(x).numel(), param_shapes))
                ),
                param_shapes,
            )
        ]
        assert len(descent_steps) == len(params)

        loss, constraint = 0.0, 0.0
        for ratio in ratios:
            for i, (step, prev_param, param) in enumerate(
                zip(descent_steps, prev_params, params)
            ):
                step = ratio * step
                new_param = prev_param - step
                params[i] = new_param

            loss = loss_fn()
            constraint = constraint_fn()
            if loss < loss_before and constraint <= max_constraint:
                break

        if not accept_violation:
            warning_msgs = []
            if loss.isnan():
                warning_msgs.append("loss is NaN")
            if constraint.isnan():
                warning_msgs.append("constraint is NaN")
            if loss >= loss_before:
                warning_msgs.append("loss is not improving")
            if constraint >= max_constraint:
                warning_msgs.append("constraint is violated")
            if warning_msgs:
                warning_msg = f"Rejecting the step: line search condition violated because {', '.join(warning_msgs)}"
                warning_msg = " and".join(warning_msg.rsplit(",", 1))
                _LOGGER.warning(warning_msg)
                for i, prev_param in enumerate(prev_params):
                    params[i] = prev_param

        return loss, constraint



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Monte Carlo return."""

from typing import Tuple

import torch
from torch import Tensor

from actorch.algorithms.value_estimation.n_step_return import n_step_return


__all__ = [
    "monte_carlo_return",
]


def monte_carlo_return(
    rewards: "Tensor",
    mask: "Tensor",
    discount: "float" = 0.99,
    standardize_advantage: "bool" = False,
    epsilon: "float" = 1e-6,
) -> "Tuple[Tensor, Tensor]":
    """Compute the Monte Carlo returns and the corresponding advantages of a trajectory.

    In the following, let `B` denote the batch size and `T` the trajectory maximum length.

    Parameters
    ----------
    rewards:
        The rewards (`r_t` in the literature), shape: ``[B, T]``.
    mask:
        The boolean tensor indicating which elements
        are valid (True) and which are not (False),
        shape: ``[B, T]``.
    discount:
        The discount factor (`gamma` in the literature).
    standardize_advantage:
        True to standardize the advantages, False otherwise.
    epsilon:
        The term added to the denominators to improve numerical stability.

    Returns
    -------
        - The Monte Carlo returns, shape: ``[B, T]``;
        - the corresponding advantages, shape: ``[B, T]``.

    References
    ----------
    .. [1] R. S. Sutton and A. G. Barto. "Reinforcement Learning: An Introduction".
           MIT Press, 1998.
           URL: http://incompleteideas.net/sutton/book/ebook/node51.html

    """
    return n_step_return(
        state_values=torch.zeros_like(mask),
        rewards=rewards,
        dones=torch.zeros_like(mask),
        mask=mask,
        discount=discount,
        num_return_steps=mask.shape[1],
        standardize_advantage=standardize_advantage,
        epsilon=epsilon,
    )



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Epsilon-greedy agent."""

from typing import Optional, Tuple, Union

import numpy as np
from gym.spaces import Space
from numpy import ndarray
from torch import device

from actorch.agents.deterministic_agent import DeterministicAgent
from actorch.registry import register
from actorch.schedules import LinearSchedule, Schedule


__all__ = [
    "EpsilonGreedyAgent",
]


@register
class EpsilonGreedyAgent(DeterministicAgent):
    """Agent that returns a random prediction with probability `epsilon`
    (annealed over time) and a deterministic one with probability `1 - epsilon`.

    """

    def __init__(
        self,
        policy: "Policy",
        observation_space: "Space",
        action_space: "Space",
        is_batched: "bool" = False,
        epsilon: "Optional[Schedule]" = None,
        device: "Optional[Union[device, str]]" = "cpu",
    ) -> "None":
        """Initialize the object.

        Parameters
        ----------
        policy:
            The policy.
        observation_space:
            The (possibly batched) observation space.
        action_space:
            The (possibly batched) action space.
        is_batched:
            True if `observation_space` and `action_space`
            are batched, False otherwise.
        epsilon:
            The epsilon annealing schedule.
            Default to ``LinearSchedule(1., 0.05, int(1e5))``.
        device:
            The device.

        """
        self.epsilon = epsilon or LinearSchedule(1.0, 0.05, int(1e5))
        super().__init__(
            policy,
            observation_space,
            action_space,
            is_batched,
            device,
        )

    # override
    def _predict(self, flat_observation: "ndarray") -> "Tuple[ndarray, ndarray]":
        epsilon = self.epsilon()
        if epsilon < 0.0 or epsilon > 1.0:
            raise ValueError(f"`epsilon` ({epsilon}) must be in the interval [0, 1]")
        self.epsilon.step()
        flat_action, log_prob = super()._predict(flat_observation)
        explore = np.random.rand(len(flat_observation)) < epsilon
        if not explore.any():
            return flat_action, log_prob
        random_flat_action = self._flat_action_space.sample()
        random_log_prob = self._flat_action_space.log_prob(random_flat_action)
        if not self._batch_size:
            # Add batch axis
            random_flat_action = random_flat_action[None]
            random_log_prob = random_log_prob[None]
        flat_action[explore] = random_flat_action[explore]
        log_prob[explore] = random_log_prob[explore]
        return flat_action, log_prob

    def __repr__(self) -> "str":
        return (
            f"{self.__class__.__name__}"
            f"(policy: {self.policy}, "
            f"observation_space: {self.observation_space}, "
            f"action_space: {self.action_space}, "
            f"is_batched: {self.is_batched}, "
            f"epsilon: {self.epsilon}, "
            f"device: {self.device})"
        )



#!/usr/bin/env python3

# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""ACTorch main script."""

from argparse import ArgumentParser, Namespace
from typing import Any

from colorama import Fore, Style, init

from actorch.version import VERSION
from actorch.visualizer import Visualize


__all__ = [
    "ACTorch",
    "main",
]


init()


_ASCII_LOGO = (
    Fore.RED
    + Style.BRIGHT
    + f"""
    _    ____ _____              _     
   / \  / ___|_   _|__  _ __ ___| |__  
  / _ \| |     | |/ _ \| '__/ __| '_ \ 
 / ___ \ |___  | | (_) | | | (__| | | |
/_/   \_\____| |_|\___/|_|  \___|_| |_|{VERSION}

"""
    + Style.RESET_ALL
)


class ACTorch:
    """ACTorch main script."""

    _SCRIPT_TYPES = {
        # "run": Run,
        "visualize": Visualize,
    }

    def main(self, args: "Namespace") -> "None":
        """Script entry point.

        Parameters
        ----------
        args:
            The command line arguments obtained from an
            `argparse.ArgumentParser` (e.g. the default
            parser returned by method `get_default_parser`).

        """
        command = getattr(args, "command", None)
        if command not in self._SCRIPT_TYPES:
            raise ValueError(
                f"`command` ({command}) must be in {list(self._SCRIPT_TYPES.keys())}"
            )
        script = self._SCRIPT_TYPES[command]()
        del args.command
        script.main(args)

    @classmethod
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        """Return a default command line argument parser for method `main`.

        Parameters
        ----------
        parser_kwargs:
            The parser initialization keyword arguments.

        Returns
        -------
            The default command line argument parser.

        """
        parser = ArgumentParser(**parser_kwargs)
        subparsers = parser.add_subparsers(
            title="command", dest="command", required=True
        )
        for command, script_cls in cls._SCRIPT_TYPES.items():
            subparsers.add_parser(
                command,
                parents=[script_cls.get_default_parser()],
                conflict_handler="resolve",
            )
        return parser


def main() -> "None":
    """ACTorch entry point."""
    try:
        print(_ASCII_LOGO)
        parser = ACTorch.get_default_parser()
        args = parser.parse_args()
        print("------------------- Start ------------------")
        actorch = ACTorch()
        actorch.main(args)
        print("------------------- Done -------------------")
    except KeyboardInterrupt:
        print("---- Exiting early (Keyboard Interrupt) ----")


if __name__ == "__main__":
    main()



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Plotter."""

import os
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from statistics import NormalDist
from typing import Any, Dict, Tuple

import numpy as np
from numpy import ndarray
from tqdm import tqdm


__all__ = [
    "Plotter",
]


class Plotter(ABC):
    """Plotter that plots progress data."""

    def plot(
        self,
        data: "Dict[str, Dict[str, Tuple[ndarray, ndarray]]]",
        output_dirpath: "str" = "plots",
        x_name: "str" = "training_iteration",
        smoothing: "float" = 0.0,
        confidence_level: "float" = 0.9545,
        opacity: "float" = 0.15,
        **kwargs: "Any",
    ) -> "None":
        """Plot progress data.

        Parameters
        ----------
        data:
            The progress data, i.e. a dict that maps series names to dicts
            that map trial names to pairs with the following values:
            - the series mean (over sampled series);
            - the series standard deviation (over sampled series).
        output_dirpath:
            The absolute or relative path to the output directory.
        x_name:
            The name of the series to plot along the x-axis.
        smoothing:
            The smoothing factor.
        confidence_level:
            The confidence level assuming normally-distributed data.
        opacity:
            The confidence interval opacity.

        Raises
        ------
        ValueError
            If an invalid argument value is provided.
        RuntimeError
            If `x_name` is not in the progress data to plot.

        """
        for key, value in {
            "smoothing": smoothing,
            "confidence_level": confidence_level,
            "opacity": opacity,
        }.items():
            if value < 0.0 or value > 1.0:
                raise ValueError(f"`{key}` ({value}) must be in the interval [0, 1]")
        output_dirpath = os.path.realpath(output_dirpath)
        if x_name not in data:
            raise RuntimeError(
                f"`x_name` ({x_name}) must be in the progress data to plot"
            )
        x_trials = data.pop(x_name)
        x_name = x_name.replace("/", "_").replace("_", " ").capitalize()
        num_stddevs = NormalDist().inv_cdf((confidence_level + 1) / 2)
        with tqdm(total=len(data)) as progress_bar:
            for y_name, y_trials in data.items():
                progress_bar.set_description(f"Plotting {y_name}")
                output_subdirpath = os.path.join(
                    output_dirpath,
                    os.path.dirname(y_name),
                )
                os.makedirs(output_subdirpath, exist_ok=True)
                y_name = y_name.replace("/", "_")
                output_filepath = os.path.join(output_subdirpath, y_name)
                traces = {}
                for trial_name, (mean, stddev) in y_trials.items():
                    x = x_trials[trial_name][0]
                    mask = np.isfinite(x) & np.isfinite(mean)
                    x, mean, stddev = x[mask], mean[mask], stddev[mask]
                    stddev[~np.isfinite(stddev)] = 0.0
                    mean = self._exp_moving_average(mean, smoothing)
                    shift = num_stddevs * np.sqrt(
                        self._exp_moving_average(stddev ** 2, smoothing)
                    )
                    traces[trial_name] = (x, mean, shift)
                self._plot_traces(
                    traces,
                    x_name,
                    y_name.replace("_", " ").capitalize(),
                    output_filepath,
                    opacity,
                    **kwargs,
                )
                progress_bar.update()

    @classmethod
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        """Return a default command line argument parser for method `plot`.

        Parameters
        ----------
        parser_kwargs:
            The parser initialization keyword arguments.

        Returns
        -------
            The default command line argument parser.

        """
        parser = ArgumentParser(**parser_kwargs)
        parser.add_argument(
            "--output-dir",
            default="plots",
            help="absolute or relative path to the output directory",
            dest="output_dirpath",
        )
        parser.add_argument(
            "-x",
            "--x-name",
            default="training_iteration",
            help="name of the series to plot along the x-axis",
        )
        parser.add_argument(
            "-s",
            "--smoothing",
            default=0.0,
            type=float,
            help="smoothing factor",
        )
        parser.add_argument(
            "-c",
            "--confidence-level",
            default=0.9545,
            type=float,
            help="confidence level assuming normally-distributed data",
        )
        parser.add_argument(
            "-o",
            "--opacity",
            default=0.15,
            type=float,
            help="confidence interval opacity",
        )
        return parser

    def _exp_moving_average(
        self, values: "ndarray", smoothing: "float" = 0.0
    ) -> "ndarray":
        # This is the same smoothing algorithm used by TensorBoard
        # (see https://github.com/tensorflow/tensorboard/blob/2.6/tensorboard/components/vz_line_chart2/line-chart.ts#L714)
        last = values[0]
        output = []
        for i, value in enumerate(values):
            if np.isfinite(value):
                last = last * smoothing + (1 - smoothing) * value
                output.append(last)
        return np.array(output)

    @abstractmethod
    def _plot_traces(
        self,
        traces: "Dict[str, Tuple[ndarray, ndarray, ndarray]]",
        x_name: "str",
        y_name: "str",
        output_filepath: "str",
        opacity: "float",
        **kwargs: "Any",
    ) -> "None":
        """Plot traces.

        Parameters
        ----------
        traces:
            The traces, i.e. a dict that maps trial names
            to triplets with the following values:
            - the series to plot along the x-axis;
            - the mean (over sampled series) of the series to plot along the y-axis;
            - the shift (over sampled series) from the mean of the series to plot along the y-axis.
        output_filepath:
            The absolute path to the output file.
        x_name:
            The name of the series to plot along the x-axis.
        y_name:
            The name of the series to plot along the y-axis.
        opacity:
            The confidence interval opacity.

        """
        raise NotImplementedError



# ==============================================================================
# Copyright 2022 Luca Della Libera. All Rights Reserved.
# ==============================================================================

"""Loader."""

import os
import re
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from collections import defaultdict
from typing import Any, Dict, Optional, Sequence, Tuple

import numpy as np
from numpy import ndarray
from tqdm import tqdm


__all__ = [
    "Loader",
]


class Loader(ABC):
    """Loader that loads progress data from progress files stored in a directory."""

    def load(
        self,
        input_dirpath: "str",
        search_pattern: "str" = ".*",
        exclude_names: "Optional[Sequence[str]]" = None,
        **kwargs: "Any",
    ) -> "Dict[str, Dict[str, Tuple[ndarray, ndarray]]]":
        """Retrieve progress files from an input directory and load the progress data.

        Parameters
        ----------
        input_dirpath:
            The absolute or relative path to the input directory.
        search_pattern:
            The regex pattern to search for in the paths to the progress files.
        exclude_names:
            The names of the data series in the progress files that must not be loaded.
            Default to ``[]``.

        Returns
        -------
            The progress data, i.e. a dict that maps series names to dicts
            that map trial names to pairs with the following values:
            - the series mean (over sampled series);
            - the series standard deviation (over sampled series).

        """
        input_dirpath = os.path.realpath(input_dirpath)
        search_regex = re.compile(search_pattern)
        exclude_names = exclude_names or []
        filepaths = sorted(
            filter(
                lambda x: search_regex.match(x),
                [
                    os.path.join(subdirpath, filename)
                    for subdirpath, _, filenames in os.walk(input_dirpath)
                    for filename in filenames
                ],
            )
        )
        all_data = defaultdict(dict)
        with tqdm(total=len(filepaths)) as progress_bar:
            for filepath in filepaths:
                head, tail = os.path.split(filepath)
                # Assume progress files are stored in a directory named after the trial
                trial_name = os.path.basename(head)
                progress_bar.set_description(
                    f"Loading {os.path.join(trial_name, tail)}"
                )
                data = self._load_data(filepath, **kwargs)
                for series_name in data:
                    if series_name in exclude_names:
                        continue
                    if series_name.endswith("/stddev"):
                        continue
                    mean = data[series_name]
                    stddev = np.zeros_like(mean)
                    if series_name.find("/mean") != -1:
                        stddev = data.get(
                            series_name.replace("/mean", "/stddev"), stddev
                        )
                    all_data[series_name][trial_name] = (mean, stddev)
                progress_bar.update()
        return all_data

    @classmethod
    def get_default_parser(cls, **parser_kwargs: "Any") -> "ArgumentParser":
        """Return a default command line argument parser for method `load`.

        Parameters
        ----------
        parser_kwargs:
            The parser initialization keyword arguments.

        Returns
        -------
            The default command line argument parser.

        """
        parser = ArgumentParser(**parser_kwargs)
        parser.add_argument(
            "input_dirpath",
            help="absolute or relative path to the input directory",
            metavar="input-dir",
        )
        parser.add_argument(
            "-p",
            "--search-pattern",
            default=".*progress.*",
            help="regex pattern to search for in the paths to the progress files",
        )
        parser.add_argument(
            "-e",
            "--exclude-names",
            nargs="+",
            default=[],
            help="names of the data series in the progress files that must not be loaded",
        )
        return parser

    @abstractmethod
    def _load_data(self, filepath: "str", **kwargs: "Any") -> "Dict[str, ndarray]":
        """Load progress data from a progress file.

        Parameters
        ----------
        filepath:
            The absolute path to the progress file.

        Returns
        -------
            The progress data, i.e. a dict that maps names of the data
            series in the progress file to the data series themselves.

        """
        raise NotImplementedError
